/* Simple.cpp

some demonstration routines for the SimpleCapi interface.

Part of the caiviar package. 

Copyright (c) 2002 MobileX AG, http://www.mobilexag.de
Copyright (c) 2002 Matthias Kramm <kramm@quiss.org>

This file is distributed under the GPL, see file COPYING for details. */

#define LINUX
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include "../src/SimpleCapi.h"

class CapiTest {
public:

    /* starts playing a sound file and waits for it's completion */
    void playthrough(SimpleCapi*capi, int line, char*name) {
        int playjob = capi->play (line, name);
        int done = 0;
        while (!done) {
            ISDNEvent event = capi->listen(line, 0);
            switch (event.type) {
            case CAPI_EVENT_DONE:
                if (event.jobid == playjob)
                    done = 1;
                break;
            }
        }
    }

    /* wait for an incoming call, play a message to him,
       put him on hold, call an operator, play a message
       to him also, then connect the two */ 
    void testforward() {
        SimpleCapi capi("SimpleCapi.properties", 0);
        int line=capi.open ("33");
        int ret, done;
        capi.acceptCalls (line);
        while(1) {
            ISDNEvent event;
            event = capi.listen(line, 0);
            if(event.type == CAPI_EVENT_INCOMING) {
                capi.answer(&event);
                break;
            }
        }
        usleep(1000000);
        playthrough(&capi, line, "youarebeingconnected.wav");

        ret = capi.forward(line, "hold");
        done = 0;
        while (!done) {
            ISDNEvent event = capi.listen(line, 0);
            if(event.type == CAPI_EVENT_DONE)
                printf("hold jobid: %d, got jobid: %d\n", ret, event.jobid);
            break;
        }

        ret = capi.call (line, "29", 60);
        usleep(1000000);
        playthrough(&capi, line, "thereisacaller.wav");

        //ret = capi.forward(line, "ect");
        ret = capi.forward(line, "3pty");
        done = 0;
        while (!done) {
            ISDNEvent event = capi.listen(line, 0);
            if(event.type == CAPI_EVENT_DONE)
                printf("3pty jobid: %d, got jobid: %d\n", ret, event.jobid);
            break;
        }

        playthrough(&capi, line, "beep.wav");

        while(1)
            capi.listen(line,1);
    }
};

int main(int argn, char*argv[]) {
    CapiTest t;
    t.testforward();
    return 0;
}

