/* CapiServer.cpp

Implements a server for exporting ivr functionality to other hosts.

Part of the caiviar package. 

Copyright (c) 2002 MobileX AG, http://www.mobilexag.de
Copyright (c) 2002 Matthias Kramm <kramm@quiss.org>

This file is distributed under the GPL, see file COPYING for details. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef LINUX
#include <pthread.h>
#include <unistd.h>
#include <signal.h>
#else
#ifndef __windows_h_included__
#define __windows_h_included__
#include <windows.h>
#endif
#endif
#include "../config.h"
#include "TextToSpeech.h"
#include "ServerSocket.h"
#include "ClientSocket.h"
#include "../src/SimpleCapi.h"
#include "string.h"

static CServerSocket* server = 0;
static Logger*log = 0;
static SimpleCapi*capi = 0;

static int port = 7347;
static char* realspeak_dictionary = 0;

#ifndef LINUX
#define strcasecmp stricmp
#endif

#ifdef LINUX
static int dofaxconversion(char*filename, char*filename_sff)
{
    char buf[1024];
    sprintf(buf, "html2ps %s > tmp.ps", filename);
    printf("%s\n", buf);
    system(buf);
    sprintf(buf, "echo quit | gs -sDEVICE=cfax -sOutputFile=%s tmp.ps", filename_sff);
    printf("%s\n", buf);
    system(buf);
    return 1;
}
#else
static int dofaxconversion(char*filename, char*filename_sff)
{
    char buf[1024];
    sprintf(buf, "copy %s %s", filename, filename_sff);
    system(buf);
    return 1;
}
#endif

static char * getnextstring(char*buffer, int&pos)
{
  int oldpos = pos;
  if(!buffer[pos])
    return 0;
  char*chr = strchr(&buffer[pos],' ');
  if(chr)
  {
    pos = chr-buffer;
    while(buffer[pos]==' ' || buffer[pos]=='\t')
    {
      buffer[pos]=0;
      pos++;
    }
    return &buffer[oldpos];
  }
  else
  {
    return &buffer[oldpos];
  }
}

#ifndef LINUX
DWORD threadID;
HANDLE thread;
DWORD WINAPI callHandler(void*data)
#else
#define Sleep(a) usleep(a*1000)
static pthread_t thread;
static void* callHandler(void*data)
#endif
{
  TextToSpeech* tts = TextToSpeech::getImplementation(log);
  tts->setParams("realspeak.dictionary", realspeak_dictionary);
  
  char buffer[8192];
  char buffer2[8192];
  CClientSocket*socket = (CClientSocket*)data;
  ISDNEvent events[256];
  int eventcount=0;
  int line;
  int ret;
  int backoff = 1;
  do {
      line = capi->open(0); // 0 = get msisdn from properties
      if(line>=0)
	  break;
      log->logf("<notice> Couldn't create a line, trying again in %d seconds\n", backoff);
      Sleep(backoff*1000);
      backoff*=2;
      if(backoff>=300) {
	  backoff=300;
      }
  } while(line<0);

  int lasteventid = 0;
  log->logf("<notice> opened line %d\n", line);

  ret = socket->myprintf("open=0\n");

  if(!ret)
  while(1)
  {
      socket->readLine(buffer, 8192);
      memcpy(buffer2,buffer, 8192);
      printf(">%s\n", buffer);
      int bufpos = 0;
      char* string = getnextstring(buffer, bufpos);
      if(!string)
	break;
      if(!strcasecmp(string, "accept"))
      {
	    capi->acceptCalls(line);
	    if(socket->print("ok\n"))
		break;
	    printf("ok\n");
      }
      else if(!strcasecmp(string, "set"))
      {
	    char*param = getnextstring(buffer, bufpos);
	    if(!strcmp(param, "msisdn")) {
		char* msisdn = getnextstring(buffer, bufpos); 
		if(!msisdn) {
		    if(socket->print("error: missing msisdn\n"))
			break;
		    printf("error: missing msisdn\n");
		    continue;
		}
		capi->close(line);
		line = capi->open(msisdn);
	    } else {
		socket->myprintf("error: unknown parameter: %s\n", param);
		printf("error: unknown parameter: %s\n", param);
		continue;
	    }
	    if(socket->print("ok\n"))
		break;
	    printf("ok\n");
      }
      else if(!strcasecmp(string, "hangup"))
      {
	    capi->hangup(line);
	    if(socket->print("ok\n"))
		break;
	    printf("ok\n");
      }
      else if(!strcasecmp(string, "call"))
      {
	    char*msisdn = getnextstring(buffer, bufpos);
	    if(!msisdn) 
	      break;
	    char*timeout= getnextstring(buffer, bufpos);
	    if(!timeout)
	      break;
	    int ret = capi->call(line, msisdn, atoi(timeout));
	    if(socket->myprintf("ok=%d\n", ret))
		break;
	    printf("ok=%d\n",ret);
      }
      else if(!strcasecmp(string, "call_nonblock"))
      {
	    char*msisdn = getnextstring(buffer, bufpos);
	    if(!msisdn) 
	      break;
	    int ret = capi->call_nonblock(line, msisdn);
	    if(socket->myprintf("ok=%d\n", ret))
		break;
	    printf("ok=%d\n",ret);
      }
      else if(!strcasecmp(string, "play"))
      {
	    char*filename = getnextstring(buffer, bufpos);
	    if(!filename) 
	      break;
	    int ret = capi->play(line, filename);
	    if(socket->myprintf("ok=%d\n",ret))
		break;
	    printf("ok=%d\n",ret);
      }
      else if(!strcasecmp(string, "speak"))
      {
	    char*chr1 = strchr((char*)buffer2, '\"');
	    if(!chr1)
		break;
	    chr1++;
	    char*chr2 = strchr(chr1, '\"');
	    if(!chr2)
		break;
	    *chr2=0;
	    unsigned char*data;
	    unsigned long size;
	    tts->text2Stream(chr1, &data, &size);
	    if(size==0) {
		if(socket->myprintf("error=\"tts failed\"\n"))
		    break;
		printf("error=\"tts failed\"\n");
		continue;
	    }
	    int ret = capi->play(line, data, size);
	    if(socket->myprintf("ok=%d\n", ret))
		break;
	    printf("ok=%d\n",ret);
      }
      else if(!strcasecmp(string, "fax"))
      {
	    char*msisdn = getnextstring(buffer, bufpos);
	    if(!msisdn) 
	      break;
	    char*filename= getnextstring(buffer, bufpos);
	    if(!filename)
	      break;
	    dofaxconversion(filename, "tmp.sff");
	    int ret = capi->fax(line, msisdn, "tmp.sff");
	    if(socket->myprintf("ok=%d\n", ret))
		break;
	    printf("ok=%d\n", ret);
      }
      else if(!strcasecmp(string, "stop"))
      {
	    char*jobid = getnextstring(buffer, bufpos);
	    if(!jobid) 
	      break;
	    capi->stop(atoi(jobid));
	    if(socket->print("ok\n"))
		break;
	    printf("ok\n");
      }
      else if(!strcasecmp(string, "put"))
      {
	    char*filename = getnextstring(buffer, bufpos);
	    if(!filename) 
	      break;
	    if(strchr(filename, '/') || strchr(filename, '\\') || strlen(filename)>=1020)
	    {
		if(socket->myprintf("error: no '/' or '\\' characters allowed in filename\n"))
		    break;
		printf("error: no '/' or '\\' characters allowed in filename\n");
		continue;
	    }

	    char*sizestr= getnextstring(buffer, bufpos);
	    if(!sizestr)
	      break;
	    int size = atoi(sizestr);
	    int pos = 0;
	    char* buffer = (char*)malloc(size);
	    int ret;
	    memset(buffer, 0xaa, size);
	    if(!buffer) {
		printf("error: out of memory\n");
		if(socket->myprintf("error: out of memory\n"))
		    break;
		continue;
	    }
	    while(pos<size)
	    {
		int len = 1024;
		if(len > size-pos)
		    len = size-pos;
		ret = socket->getBlock(&buffer[pos], len);
		if(ret<=0) break;
		pos += ret;
	    }
	    if(ret<=0) {
		break;
	    }
	    FILE*fi = fopen(filename, "wb");
	    if(!fi) {
		printf("error: Couldn't open file %s\n", filename);
		if(socket->myprintf("error: Couldn't open %s\n", filename))
		    break;
		continue;
	    }
	    fwrite(buffer, size, 1, fi);
	    fclose(fi);
	    free(buffer);

	    if(socket->myprintf("ok\n"))
		break;
	    printf("ok\n");
      }
      else if(!strcasecmp(string, "del"))
      {
	    char*filename = getnextstring(buffer, bufpos);
	    if(!filename) 
	      break;
	    if(strchr(filename, '/') || strchr(filename, '\\') || strlen(filename)>=1020)
	    {
		if(socket->myprintf("error: no '/' or '\\' characters allowed in filename\n"))
		    break;
		printf("error: no '/' or '\\' characters allowed in filename\n");
		break;
	    }

	    char buf[1024];
#ifdef LINUX
	    sprintf(buf, "rm %s", filename);
#else
	    sprintf(buf, "del %s", filename);
#endif
	    system(buf);
	    if(socket->myprintf("ok\n"))
		break;
	    printf("ok\n");
      }
      else if(!strcasecmp(string, "listen"))
      {
	    char*timeout = getnextstring(buffer, bufpos);
	    if(!timeout) 
		timeout = "0";
	    int t = atoi(timeout);
	    ISDNEvent event;
	    event = capi->listen(line, t);
	    /*if(t) 
		t++;
	       int i;
	       for (i = 1 ;i != t; i++) {
		event = capi->listen(line, 1);
		if(event.type != TIMEOUT)
		    break;
	    }*/
	    events[eventcount] = event;
	    if(socket->myprintf("msisdn=%s\n",event.msisdn))
		break;
	    printf("msisdn=%s\n", event.msisdn);
	    if(socket->myprintf("id=%d\n",eventcount))
		break;
	    printf("id=%d\n", eventcount);
	    if(socket->myprintf("type=%d\n", event.type))
		break;
	    printf("type=%d\n", event.type);
	    if(socket->myprintf("jobid=%d\n", event.jobid))
		break;
	    printf("jobid=%d\n", event.jobid);
	    if(socket->myprintf("tone=%d\n", event.tone))
		break;
	    printf("tone=%d\n", event.tone);
	    if(socket->myprintf("reason=%d\n", event.reason))
		break;
	    printf("reason=%d\n", event.reason);
	    if(event.type == CAPI_EVENT_INCOMING)
	        lasteventid = eventcount;
	    eventcount++;
	    eventcount&=256;
	    if(socket->myprintf("ok\n"))
		break;
	    printf("ok\n");
      }
      else if(!strcasecmp(string, "answer"))
      {
	    char*eventid = getnextstring(buffer, bufpos);
	    int eid = -1;
	    if(eventid && *eventid) 
	    {
		eid = atoi(eventid);
	    }
	    if(eid>=256 || eid<0)
		eid = lasteventid;
	    int ret=capi->answer(&events[eid]);
	    if(socket->myprintf("ok=%d\n", ret))
		break;
	    printf("ok=%d\n", ret);
      }
      else if(!strcasecmp(string, "forward"))
      {
	    char*msisdn = getnextstring(buffer, bufpos);
	    if(!msisdn) 
	      break;
	    int ret = capi->forward(line, msisdn);
	    if(socket->print("ok\n"))
		break;
	    printf("ok\n");
      }
      else if(!strcasecmp(string, "close"))
      {
	    if(socket->print("ok\n"))
		break;
	    printf("ok\n");
	    break;
      }
      else 
      {
	    if(socket->print("error: Unknown command\n"))
		break;
	    printf("error: Unknown command\n");
      }
  }
  capi->close(line);
  printf("closing line %d\n", line);
	 
  delete socket;
  delete tts;
  return 0;
}

#ifdef LINUX
static void sigpipe(int par)
{
    printf("sigpipe(%d)\n", par);
    fflush(stdout);
    return;
}

int exit_done = 0;
static void sigterm(int par)
{
    printf("got signal, terminating...\n", par);
    if(server) {
	CServerSocket*tmp = server;
	server = 0;
	delete tmp;
    } 
    if(capi) {
	SimpleCapi*tmp = capi;
	capi = 0;
	delete tmp;
    }
    if(log) {
	Logger*tmp = log;
	log = 0;
	delete tmp;
    }
    if(!exit_done)
    {
	exit_done = 1;
	exit(1);
    }
}
#endif

static void readline(FILE*fi, char*line, int maxlen) {
    int pos = 0;
    while(!feof(fi)) {
        if(!fread(&line[pos],1,1,fi))
            break;
        if(line[pos] == 13 || line[pos]==10)
            break;
	if(pos<maxlen-1)
	    pos++;
    }
    line[pos]=0;

    /* cut off comments */
    char*x=strchr(line,'#');
    if(x) {
        *x=0;
	pos = x-line;
    }

    /* cut off trailing whitespace */
    while(pos>=1 && (line[pos-1]==32 || line[pos-1]==9)) {
	pos--;
	line[pos]=0;
    }
}


static void readProperties(char*name)
{
    FILE*fi=fopen(name, "rb");
    char line[256];
    char* logfilename = "ivr.log";
    int fileloglevel = 5;
    int screenloglevel = 1;
    int socketloglevel = -1;
    char* logserver = 0;
    char* logport = 0;
    int properties_allcalls = 0;

    while(fi && !feof(fi))
    {
	readline(fi, line, 256);
	char*equal = strchr(line, '=');
	if(equal)
	{
	    *equal = 0;
	    equal++;
	    printf("Property: (%s) = (%s)\n", line, equal);
	    if (!strcasecmp(line, "file.loglevel"))
                fileloglevel=atoi(equal);
            else if (!strcasecmp(line, "file.name"))
                logfilename=strdup(equal);
            else if (!strcasecmp(line, "screen.loglevel"))
                screenloglevel=atoi(equal);
            else if (!strcasecmp(line, "server.loglevel"))
                socketloglevel=atoi(equal);
            else if (!strcasecmp(line, "server.name"))
                logserver=strdup(equal);
            else if (!strcasecmp(line, "server.port"))
                logport=strdup(equal);
	    else if (!strcasecmp(line,"ivr.port")) {
		int p = atoi(equal);
		if(p>0) {
		    port = p;
		} else {
		    printf("Invalid port: %s\n", equal);
		    exit(1);
		}
	    } else if(!strcasecmp(line,"realspeak.dictionary")) {
		if(*equal)
		    realspeak_dictionary = strdup(equal);
	    }
	}
    }
    log = new Logger();
    log->initLog(logfilename, fileloglevel,
	         logserver, logport, socketloglevel,
		 screenloglevel);
}

int main(int argn, char*argv[])
{
  int t;
  char _version[80];
  char * version = _version + 2;
  sprintf(_version, "%s", VERSION);
  for(t=1;t<argn;t++) {
      if(!strcmp(argv[t], "-h")) {
	  printf("-h this help text\n");
	  printf("-v display version number\n");
	  exit(0);
      } else if(!strcmp(argv[t], "-V") ||
	        !strcmp(argv[t], "-v")) {
	  printf("capiserver version %s\n", version);
	  exit(0);
      } else if(!strcmp(argv[t], "-cvs")) {
	  printf("$Id: CapiServer.cpp,v 1.21 2002/05/02 14:53:32 kramm Exp $\n");
	  exit(0);
      } else {
	  printf("Unknown argument. Try -h\n");
	  exit(1);
      }
  }
  readProperties("ivr.properties");

  log->logf("<notice> CapiServer Version %s started.\n", version);
  capi = new SimpleCapi("ivr.properties", log);

#ifdef LINUX
  signal(SIGPIPE, sigpipe);
  signal(SIGINT , sigterm);
  signal(SIGQUIT, sigterm);
  signal(SIGTERM, sigterm);
#endif
  server = new CServerSocket();

  if(!server->initSocket(port))
  {
      log->logf("<fatal> Couldn't create listening socket\n");
      return 0;
  }
  log->logf("<notice> accepting connections on port %d\n", port);
  while(1)
  {
      CClientSocket*socket = server->acceptConnection();

      if(socket) {
#ifndef LINUX
      thread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)callHandler, socket, 0, (LPDWORD)&threadID);
#else
      thread = pthread_create(&thread, 0, callHandler, socket);
#endif
      }
  }
  if(server) {
      delete server;
      server = 0;
  }
  return 0;
}

