/* RealSpeak.h

Header file for RealSpeak.cpp

Part of the caiviar package. 

Copyright (c) 2002 MobileX AG, http://www.mobilexag.de
Copyright (c) 2002 Peter Dikant <peter.dikant@mobilexag.de>

This file is distributed under the GPL, see file COPYING for details. */

#include <windows.h>
#include <ObjBase.h>
#include "ttsso_types.h"
#include "lh_ttsso.h"
#include "../src/Log.h"

#if !defined(AFX_REALSPEAK_H__E71A4D75_A35A_48AB_9BFC_ADD1BE58699E__INCLUDED_)
#define AFX_REALSPEAK_H__E71A4D75_A35A_48AB_9BFC_ADD1BE58699E__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#define TTS_TEXT_BUFFER 1024
#define TTS_OUTPUT_BUFFER 1048576 // 1 MByte output buffer

// this header will be appended to the output created by the tts system
// to create a complete wav memory structure
static unsigned char wavHeader[] = { 
	0x52, 0x49, 0x46, 0x46, 0x00, 0x00, 0x00, 0x00, // |R|I|F|F|4 byte lenght of file - 8|
	0x57, 0x41, 0x56, 0x45, 0x66, 0x6d, 0x74, 0x20, // |W|A|V|E|f|m|t| |
	0x10, 0x00, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, // |4 byte length of header|2 bytes encoding (01 = pcm)|2 bytes number of channels|
	0x40, 0x1f, 0x00, 0x00, 0x80, 0x3e, 0x00, 0x00, // |4 byte samplesrate|4 byte samplerate * bytes per sample|
	0x02, 0x00, 0x10, 0x00, 0x64, 0x61, 0x74, 0x61, // |2 byte bytes per sample|2 byte bit per sample|d|a|t|a|
	0x00, 0x00, 0x00, 0x00 };

class RealSpeak: public TextToSpeech
{
public:
	virtual void text2Stream (const char *text, unsigned char **outbuffer, unsigned long *size);
	virtual void text2File (const char *text, const char *filename);
	virtual void text2Audio (const char *text);
	virtual void setParams (const char *key, const char* value);
	RealSpeak(Logger *log, const char* engine);
	virtual ~RealSpeak();

private:
#ifndef LINUX
	static TTSRETVAL CbTtsSource(void *lpAppData, void *lpDataBuffer, U32 nBufferSize, U32 *lpDataSize);
	static VOID *CbTtsDestination(void *lpAppData, U16 nDatatype, VOID *lpData, U32 nDataSize, U32 *pnBufferSize);
	static TTSRETVAL CbTtsEventNotify(void *pAppData, void *ppBuffer, U16 nDataSize, U16 event);
	void loadDictionary(const char* dictionary);
	HTTSINSTANCE hInst;
	TTSPARM ttsParm;
	HTTSDICT userdict;
	char text[1024];
	char enginePath[256];
	unsigned char *output;
	long outputSize;
	CRITICAL_SECTION criticalSection;
#else
	pthread_mutex_t criticalSection;
#endif
};

#endif // !defined(AFX_REALSPEAK_H__E71A4D75_A35A_48AB_9BFC_ADD1BE58699E__INCLUDED_)
