/* ServerSocket.cpp

java-like implementation of sockets.

Part of the caiviar package. 

Copyright (c) 2002 MobileX AG, http://www.mobilexag.de
Copyright (c) 2002 Peter Dikant <peter.dikant@mobilexag.de>

This file is distributed under the GPL, see file COPYING for details. */

#ifdef LINUX
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <time.h>
#include <memory.h>
#include <unistd.h>
#include "socket.h"
#include <fcntl.h>
#include <sys/types.h>
#include <pthread.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <net/if.h>
#include <netdb.h>
#else
#ifndef __windows_h_included__
#define __windows_h_included__
#include <windows.h>
#endif
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#endif
#include "ServerSocket.h"


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

#ifdef LINUX
static void Sleep(int useconds)
{
	usleep(useconds);
}
static void closesocket(int socket)
{
	close(socket);
}
static const char FALSE = 0;
static const char TRUE = 1;
typedef sockaddr_in SOCKADDR_IN;
typedef sockaddr_in* LPSOCKADDR;
#endif

CServerSocket::CServerSocket() 
{

}

/**
 * close the socket and clean up the network layer before disposing this object.
 */
CServerSocket::~CServerSocket()
{
	closesocket(listeningSocket);
#ifndef LINUX
	WSACleanup();
#endif
}

/**
 * initialize the windows socket layer and start a server socket on the specified port.
 */
BOOL CServerSocket::initSocket(short port)
{
#ifndef LINUX
	WSAData wsaData;
#endif
	int nRet;

	nPort = port;

#ifndef LINUX
	// Start Winsock up
	nRet = WSAStartup(MAKEWORD(2, 2), &wsaData);
	if (nRet != 0) {
		cerr << "WSAStartup() returned error code " << nRet << "." << endl;
		return FALSE;
	}
#endif

	listeningSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
	if (listeningSocket == INVALID_SOCKET) {
#ifndef LINUX
		cerr << "Could not create listening socket: " << WSAGetLastError() << endl;
		WSACleanup();
#else
		perror("Could not create listening socket:");
#endif
		return FALSE;
	}
 
	SOCKADDR_IN sinInterface;
	sinInterface.sin_family = AF_INET;
	sinInterface.sin_addr.s_addr = INADDR_ANY;
	sinInterface.sin_port = htons(nPort);
#ifndef LINUX  
	nRet = bind(listeningSocket, (LPSOCKADDR)&sinInterface, sizeof(sinInterface));
#else
	nRet = bind(listeningSocket, (struct sockaddr*)&sinInterface, sizeof(sinInterface));
#endif
	if (nRet == SOCKET_ERROR) {
#ifndef LINUX
		cerr << "Bind error: " << WSAGetLastError() << endl;
		closesocket(listeningSocket);
		WSACleanup();
#else
		perror("bind error:");
		closesocket(listeningSocket);
#endif
		return FALSE;
	}

	nRet = listen(listeningSocket, SOMAXCONN);
	if (nRet == SOCKET_ERROR) {
#ifndef LINUX
		cerr << "Listen error: " << WSAGetLastError() << endl;
		closesocket(listeningSocket);
		WSACleanup();
#else
		perror("listen error:");
		closesocket(listeningSocket);
#endif
		return FALSE;
	}
	return TRUE;
} /* initSocket */

/**
 * this method will wait for a new incoming connection on the server port.
 */
CClientSocket* CServerSocket::acceptConnection()
{
	sockaddr_in sinRemote;
	int nAddrSize = sizeof(sinRemote);

	SOCKET client = accept(listeningSocket, (sockaddr*)&sinRemote, 
#ifdef LINUX
		(unsigned int*)
#endif
		&nAddrSize);
	if (client == INVALID_SOCKET) {
#ifndef LINUX
		cerr << "Could not accept connection: " << WSAGetLastError() << endl;
#else
		perror("Could not accept connection: ");
#endif
		return NULL;
	}
#ifdef LINUX
        fcntl (client, F_SETFL, 0);	//sock?
#endif
	
	printf("Accepted connection from %s:%d.\n",inet_ntoa(sinRemote.sin_addr),
						   ntohs(sinRemote.sin_port));

	if (client == SOCKET_ERROR) {
		return NULL;
	}
	CClientSocket* cs = new CClientSocket(client);
	cs->host = inet_ntoa(sinRemote.sin_addr);
	cs->port = ntohs(sinRemote.sin_port);
	return cs;
} /* accept */
