/* socket.c

linux implementation of sockets. Used by ClientSocket and
ServerSocket.

Part of the caiviar package. 

Copyright (c) 2002 MobileX AG, http://www.mobilexag.de
Copyright (c) 2002 Matthias Kramm <kramm@quiss.org>

This file is distributed under the GPL, see file COPYING for details. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <fcntl.h>
#include <sys/types.h>
#include <pthread.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <net/if.h>
#include <netdb.h>


int
create_listen_socket (int port)
{
  struct sockaddr_in sin;
  int sock;
  int ret, i;
  int val = 1;
  sin.sin_addr.s_addr = INADDR_ANY;
  sin.sin_port = htons (port);
  sin.sin_family = AF_INET;

  sock = socket (AF_INET, SOCK_STREAM, 6);
  if (sock < 0)
    {
      printf ("create_listen_socket, socket\n");
      exit (1);
    }

  if (setsockopt (sock, SOL_SOCKET, SO_REUSEADDR, &val, sizeof (val)) < 0)
    {
      printf ("setsockopt(REUSEADDR\n");
      exit (1);
    }

  if (bind (sock, (struct sockaddr *) &sin, sizeof (sin)) < 0)
    {
      perror ("bind");
      exit (1);
    }

  if (listen (sock, 5) < 0)
    {
      printf ("create_listen_socket, listen");
      exit (1);
    }

  if (fcntl (sock, F_SETFL, O_NONBLOCK) < 0)
    {
      printf ("fcntl(NONBLOCK)\n");
      return ret;
    }

  printf ("created socket bound to port %d\n", port);
  fflush(stdout);

  return sock;
}


int
accept_connection (int socket, int timeout)
{
  fd_set fsr;
  struct timeval tv;
  int ret;
  FD_ZERO (&fsr);
  FD_SET (socket, &fsr);
  tv.tv_sec = timeout / 1000000;
  tv.tv_usec = timeout % 1000000;
  ret = select (socket + 1, &fsr, 0, 0, &tv);
  if (FD_ISSET (socket, &fsr))
    {
      int newsock;
      struct sockaddr_in sin;
      unsigned int len;
      int ret;

      len = sizeof (sin);
      newsock = accept (socket, (struct sockaddr *) &sin, &len);

      if (newsock < 0)
	{
	  printf ("couldn't accept\n");
	  fflush(stdout);
	  return -1;
	}
      fcntl (newsock, F_SETFL, 0);	//sock?
      return newsock;
    }
  else
    return -1;
}

int
connect_to_host (char *host, unsigned int port)
{
  int i, sock = -1, ret;
  char buf_ip[100];
  struct hostent *he; 
  unsigned char ip1;
  unsigned char ip2;
  unsigned char ip3;
  unsigned char ip4;
  struct sockaddr_in sin;
  int timeout = 120;

  he = gethostbyname (host);
  if (!he)
    {
      printf ("Couldn't resolve %s\n", host);
      fflush(stdout);
      return -1;
    }

  ip1 = he->h_addr_list[0][0];
  ip2 = he->h_addr_list[0][1];
  ip3 = he->h_addr_list[0][2];
  ip4 = he->h_addr_list[0][3];
  sin.sin_family = AF_INET;
  sin.sin_port = (port << 8) + (port >> 8);
  memcpy (&sin.sin_addr.s_addr, he->h_addr_list[0], 4);

  printf ("Connecting to %d.%d.%d.%d:%d...\n", ip1, ip2, ip3, ip4, port);
  fflush(stdout);

  sock = socket (AF_INET, SOCK_STREAM, 6);
  if (sock < 0)
    {
      printf("socket (AF_INET, SOCK_STREAM, 6) failed\n");
      fflush(stdout);
      return -1;
    }

  ret = connect (sock, (struct sockaddr *) &sin, sizeof (struct sockaddr_in));
  if (ret < 0)
    {
      printf("connect (socket, sin, len) failed\n");
      fflush(stdout);
      return -1;
    }

  return sock;
}

