/* CapiHandler.cpp
 *
implementation of the CCapiHandler2 class.

Part of the caiviar package. 

Copyright (c) 2002 MobileX AG, http://www.mobilexag.de
Copyright (c) 2002 Matthias Kramm <kramm@quiss.org>

This file is distributed under the GPL, see file COPYING for details. */

#if defined(LINUX) || defined(CYGWIN)
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <memory.h>
#include <unistd.h>
#include "CapiHandler2.h"
#include "Log.h"
#include "fax.h"
#define Sleep usleep
#elif defined(WIN32)
#include "CapiHandler2.h"
#include "log.h"
#include <malloc.h>
#include <time.h>
#include "cwin32.h"
#include "fax.h"
#ifndef __windows_h_included__
#define __windows_h_included__
#include <windows.h>
#endif
#endif
#include "range.h"

// TODO: * allow stopping of call and callfax jobs
//       * make a job table
//       * make listen() a thread?

/* if you set this to 1, the capi won't send a alert_req immediately
 * after each connect_ind. This means that other applications are free
 * to take the call as long as we didn't. Unless you're sure what you're
 * doing, set this to 0 */
#define DONT_DO_ALERTING 0

// defines
//#define MaxNumBChan        2  /*----- max. number of B-channels -----*/
#define MaxNumBChan        1  /*----- max. number of B-channels -----*/
#define MaxNumB3DataBlocks 7  /*----- max. number of unconfirmed B3-datablocks -----*/
/*----- 7 is the maximal number supported by CAPI -----*/
#define MaxB3DataBlockSize 2048  /*- max. B3-Datablocksize -----*/
/*- 2048 is the maximum supported by CAPI -----*/
#define MsgBufSize  (1024 + (1024 * MaxNumBChan)) /*-- size of messagebuffer -----*/

#define SendBlockSize   2048	/*----- must not be greater than MaxB3DataBlockSize in req.c-----*/
#define READ_FORWARD 4          /* must be < 8 | bigger=more stable, smaller=better barge in */

#define TRACE_MESSAGES

static unsigned char global_job_counter = 1;

#define CONTROLLER(line) (channeltab[line].controller)
#define APPL_ID(line) (channeltab[line].Appl_Id)
#define FLAG_LISTEN(line) (channeltab[line].flag_listen)
#define FLAG_DTMF(line) (channeltab[line].flag_dtmf)
#define FLAG_FAX(line) (channeltab[line].flag_fax)
#define FLAG_X31(line) (channeltab[line].flag_x31)
#define STATE(line) (channeltab[line].state)

/* these are from ETS 300 102-1
 *
 * you should normally only see are very small subset of these causes.
 * Do not change the simplified mappings unless you're sure what you're doing.
 */
struct reason_t {
    int number;
    char*name;
    int simplified;
}
reasontab[] = { /*000 0001*/  {1, "Unallocated (unassigned) number", CAPI_REASON_WRONGNUMBER},
                              /*000 0010*/  {2, "No route to specified transit network", CAPI_REASON_TECHNICAL},
                              /*000 0011*/  {3, "No route to destination", CAPI_REASON_TECHNICAL},
                              /*000 0110*/  {6, "channel unacceptable", CAPI_REASON_TECHNICAL},
                              /*000 0111*/  {7, "Call awarded and being delivered in an established channel", CAPI_REASON_TECHNICAL},
                              /*001 0000*/ {16, "Normal call clearing", CAPI_REASON_REJECTED},
                              /*001 0001*/ {17, "User Busy", CAPI_REASON_BUSY},
                              /*001 0010*/ {18, "No User responding", CAPI_REASON_REJECTED},
                              /*001 0011*/ {19, "No answer from user (user alerted)", CAPI_REASON_REJECTED},
                              /*001 0101*/ {21, "Call rejected", CAPI_REASON_REJECTED},
                              /*001 0110*/ {22, "Number changed", CAPI_REASON_WRONGNUMBER},
                              /*001 1010*/ {26, "Nonselected user clearing", CAPI_REASON_TECHNICAL},
                              /*001 1011*/ {27, "Destination out of order", CAPI_REASON_TECHNICAL},
                              /*001 1100*/ {28, "Invalid number format", CAPI_REASON_TECHNICAL},
                              /*001 1101*/ {29, "Faculty rejected", CAPI_REASON_TECHNICAL},
                              /*001 1110*/ {30, "Response to status inqury", CAPI_REASON_TECHNICAL},
                              /*001 1111*/ {31, "Normal, unspecified", CAPI_REASON_TECHNICAL},
                              /*010 0010*/ {34, "No circuit/channel available", CAPI_REASON_BUSY},
                              /*010 0110*/ {38, "Network out of order", CAPI_REASON_TECHNICAL},
                              /*010 1001*/ {41, "Temporary failure", CAPI_REASON_TECHNICAL},
                              /*010 1010*/ {42, "Switching equipment congestion", CAPI_REASON_TECHNICAL},
                              /*010 1011*/ {43, "Access information discarded", CAPI_REASON_TECHNICAL},
                              /*010 1100*/ {44, "Requested circuit/channel not available", CAPI_REASON_TECHNICAL},
                              /*010 1111*/ {47, "Resource unavailable, unspecified", CAPI_REASON_TECHNICAL}
              };
char* supplementary_service_names[] =
    {"Get supported Services",
     "Listen",
     "Hold",
     "Retrieve",
     "Suspend",
     "Resume",
     "ECT (Explicit Call Transfer)",
     "3PTY Begin",
     "3PTY END",
     "CF Activate",
     "CF Deactivate",
     "CF Interrogate Parameters",
     "CF Interrogate Numbers",
     "CD (Call deflection)",
     "MCID (Malicious Call Identification",
     "CCBS Request (Completion of Calls to busy Sub)",
     "CCBS Deactivate",
     "CCBS Interrogate",
     "CCBS Call"
    };
int supplementary_service_num = sizeof(supplementary_service_names)/sizeof(char*);


/* Test if CAPI is installed, and furthermore return
 * the number of accessible controllers
 */
unsigned int CCapiHandler2::InitCAPI() {
    unsigned int numController;

    if (!CAPI20_ISINSTALLED()) {
        log->logf("<fatal> InitCapi: ISDN-driver not installed");
        return 0;
    }

    // get number of controllers
    unsigned short Buffer[64 / sizeof (unsigned short)];
    CAPI20_GET_PROFILE (0, (unsigned char *)Buffer);
    numController = Buffer[0];

    if (numController == 0) {
        log->logf("<fatal> InitCapi: No ISDN-controller installed");
        return 0;
    }

    return numController;
}

/* Register, open, and initialize a specific line.
 * The msisdn is the msisdn under which the line
 * should listen for calls _and_ which should appear
 * for the called side when making calls
 */
int CCapiHandler2::openLine(int line, char*msisdn) {
    if(channeltab[line].active) {
        log->logf("<warning> [line %d] Can't register line %d. Already active.", line, line);
        return 0;
    }

    CAPI_REGISTER_ERROR ErrorCode;
#if defined(WIN32) || defined(CYGWIN)
    APPL_ID(line) = CAPI20_REGISTER(MsgBufSize, MaxNumBChan, MaxNumB3DataBlocks, MaxB3DataBlockSize, &ErrorCode);
#else
    APPL_ID(line) = CAPI20_REGISTER(MaxNumBChan, MaxNumB3DataBlocks, MaxB3DataBlockSize, &ErrorCode);
#endif
    if (APPL_ID(line) == 0) {
        log->logf("<fatal> [line %d] registering the application in function RegisterCAPI, ", line);
        log->logf("<fatal> [line %d] code: %04X\n", line,ErrorCode);
        return 0;
    }
    log->logf("<verbose> [line %d] Registered as BChannel %d of Controller %d, ApplicatonID: %d",
	    line, channeltab[line].controller_line, channeltab[line].controller, 
	    APPL_ID(line));

    channeltab[line].active = 1;
    channeltab[line].my_msisdn = msisdn;
    FLAG_LISTEN(line) = 0;
    resetLineToIdle(line);
    return 1;
}

/* Reset, close and Deregister a specific line.
 */
void CCapiHandler2::close(int line) {
    MESSAGE_EXCHANGE_ERROR ErrorCode;
    if(line>=0 && line<maxchannels && channeltab[line].active) {
        if(STATE(line) == state_connected || STATE(line) == state_holdreq)
            this->hangup(line);

        if(!channeltab[line].active)
            return;
        log->logf("<notice> [line %d] Releasing line %d (Appl %d)", line, line, APPL_ID(line));

        ErrorCode = CAPI20_RELEASE(APPL_ID(line));
        if (ErrorCode != 0) {
            log->logf("<error> [line %d] deregistering the application in function close(), ", line);
            log->logf("<error> [line %d] code: 0x%04X\n", line,ErrorCode);
        }
        channeltab[line].active = 0;
    }
}

/* GetNumOfSupportedBChannels: Returns the number of supported B-channels
 * for the specified controller
 */
unsigned int CCapiHandler2::GetNumOfSupportedBChannels(long Controller) {
    unsigned short Buffer[64 / sizeof (unsigned short)];

    /*----- retrieve controller specific information -----*/
    CAPI20_GET_PROFILE((unsigned)Controller, (unsigned char *)Buffer);
    // Buffer actually contains a lot of useful information
    // we might be interested in.

    return (unsigned)Buffer[1];
}

/* Test if a line number is valid. Used in the SimpleCapi implementation
 * to catch some obvious usage-errors.
 */
int CCapiHandler2::checkLine(int line) {
    if((line<0) || (line>=maxchannels) || (channeltab[line].active==0)) {
        log->logf("<fatal> [line %d] line %d is invalid. (did you forget to reopen a closed line?)", line, line);
        return 0;
    }
    return 1;
}

/* send a message to the capi interface. This wraps the capi function
 * CAPI_PUT_CMSG to allow printing some debugging messages.
 */
unsigned int CCapiHandler2::SendCapiMessage(int line, _cmsg* pCMSG) {
#ifdef TRACE_MESSAGES
    log->logf("<debug> [line %d] send CAPI message: 0x%02x, 0x%02x, PLCI:0x%x, NCCI:0x%x NCPI:0x%x\n", line,
              pCMSG->Command, pCMSG->Subcommand, pCMSG->adr.adrPLCI, pCMSG->adr.adrNCCI, pCMSG->NCPI);
#endif
    unsigned int retval = CAPI_PUT_CMSG(pCMSG);
    if (retval != 0) {
        log->logf("<error> [line %d] Couldn't send message: 0x%02x, 0x%02x, PLCI:0x%x, NCCI:0x%x NCPI:0x%x\n", line,
                  pCMSG->Command, pCMSG->Subcommand, pCMSG->adr.adrPLCI, pCMSG->adr.adrNCCI, pCMSG->NCPI);
    }

    return retval;
}

/* Constructor. Intitialize mutex(s) and the channeltab.
 */
CCapiHandler2::CCapiHandler2(Logger*log) {
    this->log = log;
    this->allcalls_hack = 0;
    log->log ("<verbose> CCapiHandler2::CCapiHandler2()");
#if defined(WIN32)
    InitializeCriticalSection(&capi_mutex);
#endif
    init();
}

/* Destructor. close all lines that are open.
 */
CCapiHandler2::~CCapiHandler2() {
    log->log ("<verbose> CCapiHandler2::~CCapiHandler2()");
    int t;
    for(t=0;t<maxchannels;t++) {
        close(t);
    }
#if defined(WIN32)
    DeleteCriticalSection(&capi_mutex);
#endif
}

/* init(). Called from the constructor. Initializes the
 * capi interface and set's up the channeltab.
 */
int CCapiHandler2::init() {
    unsigned int numDevs = InitCAPI();
    if (numDevs == 0)
        return 0;
    log->logf("<notice> %d Controller%s found\n", numDevs, numDevs>=2?"s":"");

    unsigned int lines = 0;
    int globlines = 0;
    memset(channeltab,0,sizeof(channeltab));
    for (unsigned int i=1; i<=numDevs; i++) {
        unsigned char buf[12];

        CAPI20_GET_SERIAL_NUMBER(buf);

        lines = GetNumOfSupportedBChannels(i);
        log->logf("<notice> Controller #%d (%s) supports %d lines\n", i, buf, lines);

        unsigned int j;
        for(j=0;j<lines;j++) {
            channeltab[globlines].active = 0;
            channeltab[globlines].controller = i;
            channeltab[globlines].controller_line = j;
            channeltab[globlines].other_msisdn[0] = 0; // set msisdn to ""
            channeltab[globlines].playjob.bytes_remaining = 0;
            channeltab[globlines].recordjob.bytes_remaining = 0;
            channeltab[globlines].msgbufsize = 0;

            globlines ++;
        }
    }
    maxchannels = globlines;

    return maxchannels;
}

/* Reset a line to idle state. Notice! This just clears
 * all variables. It doesn't restore any physical states.
 */
void CCapiHandler2::resetLineToIdle(int line) {
    STATE(line) = state_idle;
    if(channeltab[line].playjob.bytes_remaining) {
        free(channeltab[line].playjob.data);
        channeltab[line].playjob.bytes_remaining = 0;
    }
    channeltab[line].recordjob.bytes_remaining = 0;
    if(channeltab[line].msgbufsize) {
        log->logf("<verbose> [line %d] Resetting line. %d unprocessed messages in messagequeue.", line,
                  channeltab[line].msgbufsize);
        channeltab[line].msgbufsize = 0;
    }
    channeltab[line].msgbufreadpos = 0;
    channeltab[line].msgbufwritepos = 0;
    FLAG_DTMF(line) = 0;
    FLAG_FAX(line) = 0;
    FLAG_X31(line) = 0;
    channeltab[line].flag_hold = 0;
    channeltab[line].flag_3pty = 0;
}

/* Open a free line, and assign it to the given msisdn.
 * If there's no free line, returns -1.
 */
int CCapiHandler2::open (char*msisdn, char*controller) {
    int t;
    for(t=0;t<maxchannels;t++) {
        if(channeltab[t].active == 0
	   && is_in_range(channeltab[t].controller, controller)) {
            if(!openLine(t, msisdn))
                return -1;

            return t;
        }
    }
    log->logf("<warning> No free line in range %s!\n", controller);
    return  -1;
}

/* acceptCalls. Notifies the capi interface that the caller want's to get notified
 * of incoming calls.
 */
void CCapiHandler2::acceptCalls(int line) {
    int ret;
    FLAG_LISTEN(line) = 1;
    // listen to incoming calls
    _cmsg mycmsg;
    LISTEN_REQ_HEADER(&mycmsg, APPL_ID(line), (unsigned short) this, CONTROLLER(line));
    LISTEN_REQ_CONTROLLER(&mycmsg) = CONTROLLER(line);
    LISTEN_REQ_INFOMASK(&mycmsg) = 128; //0;
    LISTEN_REQ_CIPMASK(&mycmsg) = 0x1FFF03FF;   // ALL_SERVICES
    LISTEN_REQ_CIPMASK2(&mycmsg) = 0;
    LISTEN_REQ_CALLINGPARTYNUMBER(&mycmsg) = NULL;
    LISTEN_REQ_CALLINGPARTYSUBADDRESS(&mycmsg) = NULL;
    log->logf("<verbose> [line %d] Sending LISTEN_REQ...", line);
    ret = SendCapiMessage(line, &mycmsg);
    if(ret) {
        log->logf("<error> [line %d] Could not send LISTEN_REQ. Trying again in 5 seconds...", line);
        Sleep(5000);
        ret = SendCapiMessage(line, &mycmsg);
        if(!ret) {
            log->logf("<fatal> [line %d] Could not send LISTEN_REQ. Listening for calls NOT enabled.", line);
            return ;
        }
    }

    log->logf("<verbose> [line %d] Waiting for LISTEN_CONF...", line);

    // wait for LISTEN_CONF
    mycmsg = waitForEvent(line, CAPI_LISTEN, CAPI_CONF, 1);

    if (LISTEN_CONF_INFO(&mycmsg) != 0) {
        log->logf ("<warning> [line %d] error at registering listener: 0x%02x\n", line, LISTEN_CONF_INFO(&mycmsg));
        return ;
    }
    log->logf("<verbose> [line %d] ...done", line);
    log->logf("<verbose> [line %d] CAPI listener registered correctly\n", line);
}

/* getCapiMessage().  Tries to get a message in the specified time.
 * The timeout variable is a reference. It is used to return 
 * how much time was "left" when we got the message.
 */
_cmsg CCapiHandler2::getCapiMessage(int line, float& timeout) {
    unsigned int mexchError;   // MESSAGE_EXCHANGE_ERROR
    _cmsg mycmsg;

    time_t start = time(0);
    float timespend;
    float inittime = timeout;

    do {
        time_t now = time(0);
        timespend = (float)difftime(now , start);
        if (timespend >= timeout && inittime>0)
            break;

        //unsigned int waitRet = CAPI20_WAIT_FOR_SIGNAL(APPL_ID(line));
        mexchError = CAPI_GET_CMSG(&mycmsg, APPL_ID(line));
        if ((MESSAGE_EXCHANGE_ERROR)mexchError != CapiNoError &&
                (MESSAGE_EXCHANGE_ERROR)mexchError != CapiReceiveQueueEmpty)
            log->logf("<error> [line %d] Message exchange error %04x\n", line, mexchError);

        if(mexchError)
            Sleep(33);
    } while (mexchError != 0);

    timeout -= timespend;
    if (timeout<=0 && inittime>0) {
        timeout = 0;
        log->logf("<verbose> [line %d] TIMEOUT", line);
        return mycmsg;
    }

    if(mycmsg.AdditionalInfo)
        log->logf("<verbose> [line %d] CAPI message has Info field set: 0x%02x, 0x%02x, (Info:%x) PLCI:0x%x, NCCI:0x%x NCPI:0x%x\n", line,
                   mycmsg.Command, mycmsg.Subcommand, mycmsg.AdditionalInfo, mycmsg.adr.adrPLCI, mycmsg.adr.adrNCCI,mycmsg.NCPI
                 );
#ifdef TRACE_MESSAGES
    else
        log->logf("<debug> [line %d] got CAPI message: 0x%02x, 0x%02x, (Info:%x) PLCI:0x%x, NCCI:0x%x NCPI:0x%x\n", line,
                   mycmsg.Command, mycmsg.Subcommand, mycmsg.AdditionalInfo, mycmsg.adr.adrPLCI, mycmsg.adr.adrNCCI,mycmsg.NCPI
                 );
#endif
    return mycmsg;
}

/* Put a (usally unprocessed) message into the message queue. This queue
 * will be read next time listen() is called.
 */
void CCapiHandler2::insertIntoQueue(int line, _cmsg mycmsg) {
    if(channeltab[line].msgbufsize == MSGBUFSIZE) {
        channeltab[line].msgbufreadpos++;
        channeltab[line].msgbufreadpos %= MSGBUFSIZE;
        channeltab[line].msgbufsize--;
    }
    channeltab[line].msgbuf[channeltab[line].msgbufreadpos] = mycmsg;
    channeltab[line].msgbufwritepos++;
    channeltab[line].msgbufwritepos %= MSGBUFSIZE;
    channeltab[line].msgbufsize++;
}

/* wait till a specified message comes in. E.g. event=0x00, subevent=0x82 can be used
 * to wait for the next IND message
 */
_cmsg CCapiHandler2::waitForEvent(int line, int event, int subevent, int store) //0 == joker!
{
    _cmsg mycmsg;
#ifdef TRACE_MESSAGES
    log->logf("<debug> [line %d] Blocking wait for 0x%02x 0x%02x", line, event, subevent);
#endif
    while (true) {
        float timeout = 1;
        mycmsg = getCapiMessage(line, timeout);
        if(timeout > 0) {
            if ((mycmsg.Command == event || event == 0 ) &&
                    (mycmsg.Subcommand == subevent || subevent == 0)) {
#ifdef TRACE_MESSAGES
                log->logf("<debug> [line %d] Blocking wait for 0x%02x 0x%02x ended.", line,event, subevent);
#endif
                break;
            } else {
                log->logf("<notice> [line %d] [wait for 0x%02x 0x%02x] unexpected reply: 0x%02x 0x%02x %s\n", line,
                          event, subevent,
                          mycmsg.Command, mycmsg.Subcommand,
                          store?"(storing)":""
                         );
                if(store)
                    insertIntoQueue(line, mycmsg);
            }
        }
    }
    return mycmsg;
}

/* set the datex address to be used for a x31 call
 * @param line the isdn line for the call
 * @param address the datexp address as a string
 */
void CCapiHandler2::setDatexAddress (int line, char* address) {
    strncpy(channeltab[line].other_datexp, address, sizeof(channeltab[line].other_datexp));
}

/* use a (already openend) line to call a specified msisdn.
 * mode < 0: normal call
 * mode == 0: fax sff
 * mode == 5: fax ascii
 * mode == 100: data x31
 */
int CCapiHandler2::call (int line, char* msisdn, int mode) {
    unsigned char b2conf[] = { 0x05, 0x03, 0x01, 0x08, 0x07, 0x00 };
    unsigned char bcconf[] = { 0x02, 0x88, 0x90 };
    unsigned int len1 = strlen((char*)msisdn);
    unsigned char* pCalledNumberStruct = (unsigned char *)malloc(len1 + 2 + 1); // gets freed
    pCalledNumberStruct[0] = (unsigned char)(len1 + 1);
    pCalledNumberStruct[1] = 0x80;
    strcpy ((char*)&(pCalledNumberStruct[2]), (char*)msisdn);

    char*mynumber = channeltab[line].my_msisdn;
    unsigned int len2 = (mynumber == NULL)? 1 : strlen((char*)mynumber);
    unsigned char* pCallingNumberStruct = (unsigned char *)malloc(len2 + 2 + 1); // gets freed
    if (mynumber != NULL) {
        pCallingNumberStruct[0] = (unsigned char)(len2 + 1);
        pCallingNumberStruct[1] = 0x80;
        strcpy ((char*)&(pCallingNumberStruct[2]), (char*)mynumber);
    } else {
        // no msn supplied, fill in basic values
        pCallingNumberStruct[0] = 0x02;
        pCallingNumberStruct[1] = 0x00;
        pCallingNumberStruct[2] = 0x80;
    }

    int b1proto = 1; // 64 kbit with byte framing
    int b2proto = 1; // Transparent
    int b3proto = 0; // Transparent

    B3_PROTO_FAXG3  B3conf;
    B3_PROTO_X31 B3confX31;
    if (mode == 100) {
        b1proto = 0; // 64 kbit with HDLC framing
        b2proto = 0; // ISO 7776
        b3proto = 2; // ISO 8208 (X.25 DTE-DTE)
        B3confX31.size = (char)(sizeof(B3confX31) - 1);
        B3confX31.lic = 0;
        B3confX31.hic = 0;
        B3confX31.ltc = 1;
        B3confX31.htc = 1;
        B3confX31.loc = 0;
        B3confX31.hoc = 0;
        B3confX31.modulomode = 8;
        B3confX31.windowsize = 2;
    } else if(mode >= 0) {
        b1proto = 4; // T.30 modem for Group 3 fax
        b2proto = 4; // T.30 modem for Group 3 fax
        b3proto = 4; // T.30 for Group 3 fax
        B3conf.resolution = 0;
        B3conf.format     = (unsigned short)mode;
        char    stationID[30] = "45";
        char    headLine [30] = "Simple Fax Page";
        len1 = strlen( stationID );
        B3conf.Infos[0] = (unsigned char)len1;
        strcpy( (char *)&B3conf.Infos[1], stationID );
        len2 = strlen( headLine );
        B3conf.Infos[len1 + 1] = (unsigned char)len2;
        strcpy( (char *)&B3conf.Infos[len1 + 2], headLine );
        B3conf.len = (unsigned char)(2 * sizeof( unsigned short ) + len1 + len2 + 2);
    }

    _cmsg msg;
    CONNECT_REQ_HEADER(&msg, APPL_ID(line), (unsigned short)this, CONTROLLER(line));
    CONNECT_REQ_B1PROTOCOL(&msg) = b1proto;
    CONNECT_REQ_B2PROTOCOL(&msg) = b2proto;
    CONNECT_REQ_B3PROTOCOL(&msg) = b3proto;
    CONNECT_REQ_CONTROLLER(&msg) = CONTROLLER(line);
    CONNECT_REQ_CIPVALUE(&msg) = (mode != 100) ? 1 : 7; // set to "packetmode" for x31
    CONNECT_REQ_CALLEDPARTYNUMBER(&msg) = pCalledNumberStruct;
    CONNECT_REQ_CALLINGPARTYNUMBER(&msg) = pCallingNumberStruct;
    if (mode == 100) {
        CONNECT_REQ_B2CONFIGURATION(&msg) = (_cstruct)&b2conf[0];
        CONNECT_REQ_B3CONFIGURATION(&msg) = (_cstruct)&B3confX31;
        CONNECT_REQ_BC(&msg) = (_cstruct) &bcconf[0];
    } else if(mode >= 0) {
        CONNECT_REQ_B3CONFIGURATION(&msg) = (_cstruct)&B3conf;
    }

    int ret=SendCapiMessage(line, &msg);

    free(pCallingNumberStruct);
    free(pCalledNumberStruct);
    if(!ret) {
        // wait for CONNECT_CONF
        waitForEvent (line, CAPI_CONNECT, CAPI_CONF, 1);

        STATE(line) = state_calling;
        FLAG_FAX(line) = ((mode>=0) && (mode < 100))? 1 : 0;
        FLAG_X31(line) = (mode == 100) ? 1 : 0;

        strncpy(channeltab[line].other_msisdn, msisdn, sizeof(channeltab[line].other_msisdn));
        channeltab[line].callingjob = 0x0300 | line;
        return channeltab[line].callingjob;
    } else {
        log->logf("<error> [line %d] couldn't send CONNECT_REQ: reason:%d\n", line,ret);
        STATE(line) = state_idle;
        free(pCallingNumberStruct);
        free(pCalledNumberStruct);
        return -1;
    }
}

/* find out which sound file is corrently playing, send the next datablock
 * to capi, and update the playjob structure 
 */
void CCapiHandler2::transferData(int line) {
    _cmsg  msg;

    log->logf("<verbose> [line %d] CCapiHandler2::transferData()", line);
    if (channeltab[line].playjob.bytes_remaining > 0) {
        int size = SendBlockSize;
        if(size > channeltab[line].playjob.bytes_remaining)
            size = channeltab[line].playjob.bytes_remaining;

        log->logf("<verbose> [line %d] Sending datahandle %d\n", line, channeltab[line].playjob.datahandle);

        DATA_B3_REQ_HEADER(&msg, APPL_ID(line), 1234, CONTROLLER(line));
        msg.adr.adrNCCI = channeltab[line].ncci;
#if defined(WIN32) || defined(CYGWIN)
        DATA_B3_REQ_DATA(&msg) = (unsigned long)channeltab[line].playjob.data_pos;
#else
        DATA_B3_REQ_DATA(&msg) = (unsigned char*)(unsigned long)channeltab[line].playjob.data_pos;
#endif
        DATA_B3_REQ_DATALENGTH(&msg) = size;
        if (FLAG_X31(line) == 0) {
            DATA_B3_REQ_DATAHANDLE(&msg) = channeltab[line].playjob.datahandle;
            DATA_B3_REQ_FLAGS(&msg) |= 2; //more data //new
        } else {
            DATA_B3_REQ_DATAHANDLE(&msg) = 1;
            DATA_B3_REQ_FLAGS(&msg) = 0;
        }

        if(!SendCapiMessage(line, &msg)) {
            channeltab[line].playjob.data_pos += size;
            channeltab[line].playjob.bytes_remaining -= size;
            if(channeltab[line].playjob.bytes_remaining == 0) {
                free(channeltab[line].playjob.data);
            } else
                channeltab[line].playjob.datahandle++;
        }
    }
}

/* called from SimpleCapi- starts playing a sound file
 * (must be a .law)
 */
int CCapiHandler2::play (int line, void*mem, int len) {
    if(channeltab[line].playjob.bytes_remaining) {
        log->logf("<error> [line %d] play: Device busy. (Stop the other job first)", line);
        return 0;
    }
    int jobnr = global_job_counter++;
    channeltab[line].playjob.data = (char*)malloc (len);
    memcpy (channeltab[line].playjob.data, mem , len);
    channeltab[line].playjob.data_pos = channeltab[line].playjob.data;
    channeltab[line].playjob.data_size = len;
    channeltab[line].playjob.bytes_remaining = len;
    channeltab[line].playjob.datahandle = 256*jobnr;

    jobnr |= 0x0100;

    channeltab[line].playjob.jobid = jobnr;

    int t;
    if (FLAG_X31(line) == 0) {
        for(t=0;t<READ_FORWARD;t++)
            transferData(line);
    } else {
    	transferData(line);
    }

    return jobnr;
}

/* Tries to record the datablocks that come in all the
 * time. I sure hope they're audio data.
 */
int CCapiHandler2::record (int line, void*buf, int bufsize) {
    if(channeltab[line].recordjob.bytes_remaining) {
        log->logf("<error> [line %d] record: Device busy. (Stop the other job first)", line);
        return 0;
    }
    int jobnr = (global_job_counter++)&255;
    channeltab[line].recordjob.data = (char*)buf;
    channeltab[line].recordjob.data_pos = (char*)buf;
    channeltab[line].recordjob.data_size = 0;
    channeltab[line].recordjob.bytes_remaining = bufsize;
    channeltab[line].recordjob.datahandle = 256*jobnr;
    jobnr |= 0x0200;
    channeltab[line].recordjob.jobid = jobnr;
    return jobnr;
}

/* will report the number of bytes received over isdn in the 
 * last record attempt
 */
int CCapiHandler2::getRecordSize(int line) {
    return channeltab[line].recordjob.data_size;
}

/* will attempt to at least stop play jobs and
 * record jobs. (whether or not to stop call jobs
 * is a tta. (to think about))
 */
void CCapiHandler2::stop (int jobid) {
    if((jobid&0xff00)==0x0100) //playjob
    {
        int t;
        int line = -1;
        for(t=0;t<maxchannels;t++) {
            if(channeltab[t].playjob.jobid == jobid)
                line =t;
        }
        if( line == -1) // not found
            return;
        if(channeltab[line].playjob.bytes_remaining) {
            free(channeltab[line].playjob.data);
            channeltab[line].playjob.bytes_remaining = 0;
            channeltab[line].playjob.jobid = 0;
        }
    } else if((jobid&0xff00)==0x0200) //recordjob
    {
        int t;
        int line = -1;
        for(t=0;t<maxchannels;t++) {
            if(channeltab[t].recordjob.jobid == jobid)
                line = t;
        }
        if(line == -1) // not found
            return;
        if(channeltab[line].recordjob.bytes_remaining) {
            channeltab[line].recordjob.bytes_remaining = 0;
            channeltab[line].recordjob.jobid = 0;
        }
    } else if((jobid&0xff00)==0x0300) //call job
    {
        // FIXME
    }
}

/* capi gives us a reason when a call didn't make
 * it through. this routine resolves the
 * reason code, print out the verbose meaning,
 * and returns a simplified version
 */
int CCapiHandler2::translatereason(int line, int reason) {
    int simplified = 0;
    if(reason >= 0x3301 && reason <= 0x3303) {
        log->logf("<verbose> [line %d] Protocol error %04x", line, reason);
        simplified = CAPI_REASON_TECHNICAL;
    } else if(reason == 0x3304) {
        log->logf("<verbose> [line %d] Another application got that call", line);
        simplified = CAPI_REASON_TECHNICAL;
    } else if((reason & 0xff00) == 0x3400) {
        int t;
        int s=-1;
        for(t=0;t<(sizeof(reasontab)/sizeof(reason_t));t++) {
            if(reasontab[t].number == (reason & 0x7f))
                s=t;
        }
        if(s>=0) {
            simplified = reasontab[s].simplified;
            log->logf("<verbose> [line %d] reason:%s", line,reasontab[s].name);
        } else {
            simplified = CAPI_REASON_TECHNICAL;
            log->logf("<verbose> [line %d] unknown 34xx reason: %04x", line,reason);
        }
    } else {
        simplified = CAPI_REASON_TECHNICAL;
        log->logf("<verbose> [line %d] unknown reason: %04x", line,reason);
    }
    return simplified;
}

/* supplementary services have their own reason codings.
 * This routine just tries to translate them into english
 * and log them. */
void CCapiHandler2::showSupplementaryReason(int line, short int reason) {
    log->logf("<verbose> [line %d] supplementary service reason: %04x", line, reason);
    if(reason == 0x3303)
        log->logf("<verbose> [line %d] (network did not answer in time)", line);
}

/* handles the _cmsg, in that it sends RESP messages, performs state
 * transitions, and, if an ISDNEvent occured, copies it to the given
 * pointer  and returns 1 */
int CCapiHandler2::handleMessage(int line, _cmsg mycmsg, ISDNEvent*event) {
    log->logf("<verbose> [line %d] handling message 0x%02x 0x%02x\n", line, mycmsg.Command, mycmsg.Subcommand);

    if (IS_CONNECT_IND(&mycmsg)) {
        if( STATE(line) == state_connected) {
            // we are busy. we ignore this call, maybe some other
            // application wants to take it.
            //
            // TODO: should we send a ignore-call CONNECT_RESP here?

	    // <?
            CONNECT_RESP_HEADER(&mycmsg, APPL_ID(line), mycmsg.Messagenumber, CONTROLLER(line));
            CONNECT_RESP_PLCI(&mycmsg) = CONNECT_IND_PLCI(&mycmsg);
            CONNECT_RESP_REJECT(&mycmsg) = 1;  // 1: ignore call
            CONNECT_RESP_B1PROTOCOL(&mycmsg) = 1;
            CONNECT_RESP_B2PROTOCOL(&mycmsg) = 1;
            CONNECT_RESP_B3PROTOCOL(&mycmsg) = 0;
            this->SendCapiMessage(line, &mycmsg);
	    log->logf("<verbose> [line %d] sending CONNECT_RESP-ignore\n", line);
	    // ?>

	    //log->logf("<verbose> [line %d] ignoring CONNECT_IND\n", line);
            return 0;
        }
        if (!FLAG_LISTEN(line)) // not awaiting calls
        {
            log->log ("<notice> rejecting call\n");
            // send CONNECT_RESP
            CONNECT_RESP_HEADER(&mycmsg, APPL_ID(line), mycmsg.Messagenumber, CONTROLLER(line));
            CONNECT_RESP_PLCI(&mycmsg) = CONNECT_IND_PLCI(&mycmsg);
            CONNECT_RESP_REJECT(&mycmsg) = 2;  // 2: reject call, normal call clearing
            CONNECT_RESP_B1PROTOCOL(&mycmsg) = 1;
            CONNECT_RESP_B2PROTOCOL(&mycmsg) = 1;
            CONNECT_RESP_B3PROTOCOL(&mycmsg) = 0;
            this->SendCapiMessage(line, &mycmsg);
            return 0;
        } else {
            if(DONT_DO_ALERTING) {
                channeltab[line].plci = CONNECT_IND_PLCI(&mycmsg);
                channeltab[line].ncci = mycmsg.adr.adrNCCI;
                channeltab[line].incoming = (_cmsg*)malloc(sizeof(mycmsg));
                memcpy(channeltab[line].incoming, (void*)&mycmsg, sizeof(mycmsg));
                STATE(line) = state_alerted;
                event->line = line;
                event->type = CAPI_EVENT_INCOMING;
                event->msisdn = "";
                event->reason = 0;
                event->internal = channeltab[line].incoming;
                // we are being very friendly here. Actually the caller
                // should know these fields are not set.
                event->jobid = -1;
                event->tone = -1;
                /*return*/
                return 1;
            }

            unsigned char *number = CONNECT_IND_CALLEDPARTYNUMBER(&mycmsg);
            char * called =  (char*)malloc(number[0]+1);
            if(number[0]>=2) {
                strncpy(called, (const char*)(&number[2]), number[0]-1);
		called[number[0]-1] = 0;
		log->logf("<verbose> [line %d] incoming call to %s", line, called);
	    } else {
		called[0] = 0;
		log->logf("<verbose> [line %d] incoming call to ?", line);
	    }

	    if(!allcalls_hack)
	    if(number[0]>1 && strcmp(called, channeltab[line].my_msisdn)) {
		CONNECT_RESP_HEADER(&mycmsg, APPL_ID(line), mycmsg.Messagenumber, CONTROLLER(line));
		CONNECT_RESP_PLCI(&mycmsg) = CONNECT_IND_PLCI(&mycmsg);
		CONNECT_RESP_REJECT(&mycmsg) = 1;  // 1: ignore call
		CONNECT_RESP_B1PROTOCOL(&mycmsg) = 1;
		CONNECT_RESP_B2PROTOCOL(&mycmsg) = 1;
		CONNECT_RESP_B3PROTOCOL(&mycmsg) = 0;
		this->SendCapiMessage(line, &mycmsg);
		log->logf("<verbose> [line %d] sending CONNECT_RESP(ignore) (wrong number: (%d/%02x/\"%s\"{%d} != \"%s\"{%d}))\n", line, 
			number[0],
			number[1], 
			called, strlen(called),
			channeltab[line].my_msisdn, strlen(channeltab[line].my_msisdn));
		free(called);
		return 0;
	    }

            // send ALERT_REQ
            _cmsg resp;
            ALERT_REQ_HEADER(&resp,mycmsg.ApplId, mycmsg.Messagenumber, CONTROLLER(line));
            ALERT_REQ_PLCI(&resp) = CONNECT_IND_PLCI(&mycmsg);
            this->SendCapiMessage(line, &resp);
            STATE(line) = state_alertreq;

            char * calling = (char*)malloc(CONNECT_IND_CALLINGPARTYNUMBER(&mycmsg)[0]+1);
            if(CONNECT_IND_CALLINGPARTYNUMBER(&mycmsg)[0] <= 2)
                calling[0] = 0;
            else
                sprintf (calling, "%s", &CONNECT_IND_CALLINGPARTYNUMBER(&mycmsg)[3]);
            strncpy(channeltab[line].other_msisdn, (const char*)calling, sizeof(channeltab[line].other_msisdn));

            log->logf("<notice> [line %d] %s calling %s- connecting...\n", line, calling, called);
            channeltab[line].incoming = (_cmsg*)malloc(sizeof(mycmsg));
            memcpy(channeltab[line].incoming, (void*)&mycmsg, sizeof(mycmsg));

            unsigned char * calledsub = CONNECT_IND_CALLEDPARTYSUBADDRESS(&mycmsg);
            unsigned char *callingsub = CONNECT_IND_CALLINGPARTYSUBADDRESS(&mycmsg);

            free(calling);
            free(called);
            return 0;
        }
    } else if (IS_ALERT_CONF(&mycmsg)) {
        channeltab[line].plci = ALERT_CONF_PLCI(&mycmsg);
        channeltab[line].ncci = mycmsg.adr.adrNCCI; //FIXME
        if (STATE(line) == state_alertreq) {
            STATE(line) = state_alerted;
            log->logf("<verbose> [line %d] got ALERT_CONF\n", line);
            // we got the call into alert state. now
            // it's time to report the fact that someone is calling
            // back to our caller.
            event->line = line;
            event->type = CAPI_EVENT_INCOMING;
            event->msisdn = channeltab[line].other_msisdn;
            event->reason = 0;
            event->internal = channeltab[line].incoming;

            // we are being very friendly here. Actually the caller
            // should know these fields are not set.
            event->jobid = -1;
            event->tone = -1;
            /*return*/
            return 1;

        } else {
            log->logf("<error> [line %d] got unexpected ALERT_CONF\n", line);
            return 0;
        }
    } else if (IS_CONNECT_ACTIVE_IND(&mycmsg)) {
        log->logf("<verbose> [line %d] got connect active ind", line);
        _cmsg resp;
        // send CONNECT_ACTIVE_RESP
        CONNECT_ACTIVE_RESP_HEADER(&resp, mycmsg.ApplId, mycmsg.Messagenumber, CONTROLLER(line));
        CONNECT_ACTIVE_RESP_PLCI(&resp) = CONNECT_ACTIVE_IND_PLCI(&mycmsg);
        SendCapiMessage(line, &resp);
        // send CONNECT_B3_REQ
        CONNECT_B3_REQ_HEADER(&resp, mycmsg.ApplId, mycmsg.Messagenumber, CONTROLLER(line));
        CONNECT_B3_REQ_PLCI(&resp) = CONNECT_ACTIVE_IND_PLCI(&mycmsg);
        if (FLAG_X31(line) == 0) {
            CONNECT_B3_REQ_NCPI(&resp) = NULL;
        } else {
            unsigned char ncpi[40];
            fillNCPIStruct(line, ncpi);
            CONNECT_B3_REQ_NCPI(&resp) = (_cstruct) &ncpi[0];
        }
        SendCapiMessage(line, &resp);
        //channeltab[line].plci = CONNECT_ACTIVE_IND_PLCI(&mycmsg);
        return 0;
    } else if (IS_CONNECT_B3_ACTIVE_IND(&mycmsg)) {
        log->logf("<verbose> [line %d] got connect b3 active ind", line);
        _cmsg resp;
        // send CONNECT_B3_ACTIVE_RESP
        printf("sending b3 active response\n");
        CONNECT_B3_ACTIVE_RESP_HEADER(&resp, mycmsg.ApplId, mycmsg.Messagenumber, CONTROLLER(line));
        CONNECT_B3_ACTIVE_RESP_NCCI(&resp) = CONNECT_B3_ACTIVE_IND_NCCI(&mycmsg);
        SendCapiMessage(line, &resp);

        channeltab[line].ncci = CONNECT_B3_ACTIVE_IND_NCCI(&mycmsg);
        channeltab[line].plci = mycmsg.adr.adrPLCI;
        log->logf("<verbose> [line %d] Setting NCCI to %x", line, channeltab[line].ncci);
        log->logf("<verbose> [line %d] Setting PLCI to %x", line, channeltab[line].plci);

        // try to enable dialtones
        if(!channeltab[line].flag_fax && !channeltab[line].flag_x31) {
            FACILITY_REQ_HEADER(&resp, mycmsg.ApplId, mycmsg.Messagenumber, CONTROLLER(line));
            //FACILITY_REQ_CONTROLLER(&resp) = channeltab[line].controller;//new
            FACILITY_REQ_NCCI(&resp) = CONNECT_B3_ACTIVE_IND_NCCI(&mycmsg);
            FACILITY_REQ_FACILITYSELECTOR(&resp) = (unsigned short)1; // 1: DTMF
            /*len  listen  msdigit mspause tosend  characters */
            unsigned char cBuffer[] = {8,  1,0,    40,0,   40,0,   0,      0,         0};
            FACILITY_REQ_FACILITYREQUESTPARAMETER(&resp) = cBuffer;
            SendCapiMessage(line, &resp);
        }

        STATE(line) = state_connected;

        event->line = line;
        event->type = CAPI_EVENT_CONNECTED;
        event->msisdn = channeltab[line].other_msisdn;
        event->reason = 0;
        event->jobid = channeltab[line].callingjob;
        event->tone = -1;
        /*return*/
        return 1;
    } else if (IS_CONNECT_B3_CONF(&mycmsg)) {
        log->logf("<verbose> [line %d] got connect b3 conf", line);
        return 0;
    } else if (IS_FACILITY_CONF(&mycmsg)) {
        unsigned char*par = FACILITY_CONF_FACILITYCONFIRMATIONPARAMETER(&mycmsg);

        if (FACILITY_CONF_FACILITYSELECTOR(&mycmsg) == 3) {
            log->logf("<verbose> [line %d] got facility conf. (selector = 3)", line);
            int len0=par[0];

            int t;
            for(t=0;t<=len0;t++)
                printf("%02x ",par[t]);
            printf("\n");
            if(len0>3 && par[1]==0 && par[2]==0) { //get supplementary services
                int len=par[3];
                if(len>2 && par[4]==0 && par[5]==0) {
                    log->logf("<verbose> [line %d] supported services: %04x", line,*(unsigned long int*)(par+6));
                }
            }
        } else if (FACILITY_CONF_FACILITYSELECTOR(&mycmsg) == 1) {
            FLAG_DTMF(line) = 1;
        } else {
            log->logf("<error> [line %d] unknown facility_conf: %d (dialtones not supported?)", line,
                      FACILITY_CONF_FACILITYSELECTOR(&mycmsg));
        }
    } else if (IS_FACILITY_IND(&mycmsg)) {
        int selector = FACILITY_IND_FACILITYSELECTOR(&mycmsg);
        if (selector == 1) //DTMF
        {
            _cstruct indParams = FACILITY_IND_FACILITYINDICATIONPARAMETER(&mycmsg);
            // FIXME: if there's more than one, put them in a loop.
            int len = indParams[0];
            if(len>1)
                log->logf("<error> [line %d] missed %d dtmf(s)! (sniff)", line, indParams[0]-1);
            for(int i = 1; i <= len; i++) {
                log->logf ( "<notice> detected DTMF %d (of %d) is %c", i, len, indParams[i]);
            }
            if(len>0) {
                event->line = line;
                event->type = CAPI_EVENT_DIALTONE;
                event->msisdn = channeltab[line].other_msisdn;
                event->reason = 0;
                event->jobid = -1;
                event->tone = indParams[1];
                /*return*/
                return 1;
            }
        } else {
            unsigned char*fi = FACILITY_IND_FACILITYINDICATIONPARAMETER(&mycmsg);
            log->logf("<verbose> [line %d] Got facility ind (selector = %d)", line, selector);
            int len=fi[0];

            _cmsg resp;
            FACILITY_RESP_HEADER(&resp, mycmsg.ApplId, mycmsg.Messagenumber, CONTROLLER(line));
            FACILITY_RESP_NCCI(&resp) = FACILITY_IND_NCCI(&mycmsg);
            FACILITY_RESP_FACILITYSELECTOR(&resp) = FACILITY_IND_FACILITYSELECTOR(&mycmsg);
            //
            FACILITY_RESP_FACILITYRESPONSEPARAMETERS(&resp) = FACILITY_IND_FACILITYINDICATIONPARAMETER(&mycmsg);
            SendCapiMessage(line, &resp);
            /*	   for(t=0;t<len;t++)
            	     printf("%02x ",fi[t+1]);
            	   printf("\n");*/
            if(len>=3) {
                int function = *(short int*)&fi[1];
                int len2 = fi[3];
                int reason = -1;
                /*for(t=0;t<len2;t++)
                  printf("%02x ",fi[t+4]);
                printf("\n");*/
                if(len2>=2) {
                    reason = *(short int*)&fi[4];
                    showSupplementaryReason(line, reason);
                }

                if(function < supplementary_service_num)
                    log->logf("<verbose> [line %d] Indicator is for Function 0x%02x (%s)", line, function,
                              supplementary_service_names[function]);
                else
                    log->logf("<error> [line %d] Indicator is for Function 0x%02x (unknown service)", line, function);

                if(!reason)
                    switch(function) {
                    case 0x0d: // CD
                        break;
                    case 0x09: // CF
                        if(len2>=6) {
                            int reason = *(short int*)&fi[4];
                            int handle = *(long int*)&fi[6];
                            log->logf("<verbose> [line %d] supplementary service handle: %08x", line, handle);
                        }
                        break;
                    case 0x02: //HOLD
                        event->line = line;
                        event->type = CAPI_EVENT_DONE;
                        event->msisdn = channeltab[line].other_msisdn;
                        event->reason = 0;
                        event->jobid = channeltab[line].supplementaryjob;
                        event->tone = -1;
                        if(channeltab[line].flag_hold)
                            log->logf("<error> [line %d] Line is already holding, but got HOLD_IND", line);
                        resetLineToIdle(line);
                        channeltab[line].flag_hold = mycmsg.adr.adrPLCI;
                        return 1;
                        break;
                    case 0x03: //RETRIEVE

                        // send CONNECT_B3_REQ
                        CONNECT_B3_REQ_HEADER(&resp, mycmsg.ApplId, mycmsg.Messagenumber, CONTROLLER(line));
                        CONNECT_B3_REQ_PLCI(&resp) = mycmsg.adr.adrPLCI;
                        CONNECT_B3_REQ_NCPI(&resp) = mycmsg.NCPI;
                        SendCapiMessage(line, &resp);

                        event->line = line;
                        event->type = CAPI_EVENT_DONE;
                        event->msisdn = channeltab[line].other_msisdn;
                        event->reason = 0;
                        event->jobid = channeltab[line].supplementaryjob;
                        event->tone = -1;
                        channeltab[line].flag_hold = 0;
                        // the state "connected" will be set by CONNECT_B3_CONF, anyway
                        channeltab[line].state = state_connected;
                        return 1;
                        break;
                    case 0x07: //3PTY
                        event->line = line;
                        event->type = CAPI_EVENT_DONE;
                        event->msisdn = channeltab[line].other_msisdn;
                        event->reason = 0;
                        event->jobid = channeltab[line].supplementaryjob;
                        event->tone = -1;
                        channeltab[line].flag_hold = 0;
                        channeltab[line].flag_3pty = 1;
                        channeltab[line].state = state_connected;
                        return 1;
                        break;
                    case 0x08: //3PTY END
                        break;
                    }
            }
        }
    } else if (IS_DATA_B3_CONF(&mycmsg)) {
        int datahandle = mycmsg.DataHandle;
        if (channeltab[line].playjob.bytes_remaining) {	//still sending
            log->logf("<debug> [line %d] playjob %d datahandle %d read_forward %d", line,
                      channeltab[line].playjob.datahandle , datahandle , READ_FORWARD);
            if( channeltab[line].playjob.datahandle - datahandle < READ_FORWARD)
                transferData(line);
        } else { //all data transferred, handle remaining confirmations.
            if(datahandle == channeltab[line].playjob.datahandle) { // last block got confirmed. we're done.
                event->line = line;
                event->type = CAPI_EVENT_DONE;
                event->msisdn = channeltab[line].other_msisdn;
                event->reason = 0;
                event->jobid = channeltab[line].playjob.jobid;
                event->tone = -1;
                /*return*/
                return 1;
            }
        }
    } else if (IS_DISCONNECT_B3_CONF(&mycmsg)) {
        log->logf("<verbose> [line %d] got disconnect b3 conf", line);
    } else if (IS_DISCONNECT_CONF(&mycmsg)) {
        log->logf("<verbose> [line %d] got disconnect conf", line);
    } else if (IS_DISCONNECT_B3_IND(&mycmsg)) {
        _cmsg resp;
        DISCONNECT_B3_RESP_HEADER(&resp, mycmsg.ApplId, mycmsg.Messagenumber, CONTROLLER(line));
        DISCONNECT_B3_RESP_NCCI(&resp) = DISCONNECT_B3_IND_NCCI(&mycmsg);
        this->SendCapiMessage(line, &resp);
        if(STATE(line) == state_connected || STATE(line) == state_holdreq) {
            STATE(line) = state_disconnecting;
            if(channeltab[line].playjob.bytes_remaining)
                free(channeltab[line].playjob.data);
            channeltab[line].playjob.bytes_remaining = 0;
            channeltab[line].recordjob.bytes_remaining = 0;
        }
    } else if (IS_DISCONNECT_IND(&mycmsg)) {
	/* we always send a response, even if the message is really for
	   some other application. I think Capi expects us to. -mk */
        _cmsg resp;
        DISCONNECT_RESP_HEADER(&resp, mycmsg.ApplId, mycmsg.Messagenumber, CONTROLLER(line));
        DISCONNECT_RESP_PLCI(&resp) = DISCONNECT_IND_PLCI(&mycmsg);
        this->SendCapiMessage(line, &resp);

	/* we have to make sure it's for us */
	if(STATE(line) != state_idle && /* idle lines don't have a valid plci */
	   ((DISCONNECT_IND_PLCI(&mycmsg)&0xff00) !=
	    (channeltab[line].plci&0xff00))) {
	    log->logf("<debug> [line %d] ignoring DISCONNECT_IND (%08x!=%08x)", line,
		DISCONNECT_IND_PLCI(&mycmsg), channeltab[line].plci
	    );
	    return 0; /* not our plci - these rampant disconnect messages
			 are *very* annoying */
	}

        int reason = DISCONNECT_IND_REASON(&mycmsg);
        if(STATE (line) == state_disconnecting) {
            log->log ("<notice> returning HANGUP Event (terminating a connection)");
            log->logf ("<verbose> hangup reason: %04x", reason);
            event->line = line;
            event->type = CAPI_EVENT_HANGUP;
            event->msisdn = channeltab[line].other_msisdn;
            event->jobid = -1;
            event->tone = -1;
            event->reason = translatereason(line, reason);

            /*STATE(line) = state_idle;
            channeltab[line].playjob.bytes_remaining = 0;
            channeltab[line].recordjob.bytes_remaining = 0;*/
            resetLineToIdle(line);
            /*return*/
            return 1;
        } else if(STATE(line) == state_calling) {
            log->logf("<notice> [line %d] returning HANGUP Event (terminating a call job)", line);
            log->logf("<verbose> [line %d] reason: %04x", line, reason);

            event->line = line;
            event->type = CAPI_EVENT_HANGUP;
            event->msisdn = channeltab[line].other_msisdn;
            event->jobid = channeltab[line].callingjob;
            event->tone = -1;
            event->reason = translatereason(line, reason);

            resetLineToIdle(line);
            /*return*/
            return 1;
        } else if(STATE(line) == state_retrievereq ||
                  STATE(line) == state_holdreq ||
                  STATE(line) == state_3ptyreq) {
            log->logf("<notice> [line %d] returning HANGUP Event (terminating a suplementary job)", line);
            log->logf("<verbose> [line %d] reason: %04x", line, reason);

            event->line = line;
            event->type = CAPI_EVENT_HANGUP;
            event->msisdn = channeltab[line].other_msisdn;
            event->jobid = channeltab[line].supplementaryjob;
            event->tone = -1;
            event->reason = translatereason(line, reason);

            resetLineToIdle(line);
            /*return*/
            return 1;
        } else {
            resetLineToIdle(line);
        }
    } else if (IS_DATA_B3_IND(&mycmsg)) {

        int finished=0;
        if( channeltab[line].recordjob.bytes_remaining>0 ) {
            int size = DATA_B3_REQ_DATALENGTH(&mycmsg);
            if(size > channeltab[line].recordjob.bytes_remaining)
                size = channeltab[line].recordjob.bytes_remaining;
            log->logf("<notice> [line %d] recording - left:%d", line, channeltab[line].recordjob.bytes_remaining);
            memcpy(channeltab[line].recordjob.data_pos, (void*)DATA_B3_IND_DATA(&mycmsg), size);
            channeltab[line].recordjob.data_size += size;
            channeltab[line].recordjob.data_pos += size;
            channeltab[line].recordjob.bytes_remaining -= size;
            if(channeltab[line].recordjob.bytes_remaining == 0 && size>0) {
                finished = 1;
            } else if (FLAG_X31(line) == 1) {
            	finished = 1; // X31 packets are received at once
            }
        }

        _cmsg resp;
        DATA_B3_RESP_HEADER(&resp, mycmsg.ApplId, mycmsg.Messagenumber, CONTROLLER(line));
        DATA_B3_RESP_NCCI(&resp) = DATA_B3_IND_NCCI(&mycmsg);
        DATA_B3_RESP_DATAHANDLE(&resp) = DATA_B3_IND_DATAHANDLE(&mycmsg);
        this->SendCapiMessage(line, &resp);

        if (finished) {
            event->line = line;
            event->type = CAPI_EVENT_DONE;
            event->msisdn = channeltab[line].other_msisdn;
            event->reason = 0;
            event->jobid = channeltab[line].recordjob.jobid;
            event->tone = -1;
            /*return*/
            return 1;
        }
    }
    /*else if (IS_MANUFACTURER_IND(&mycmsg)) {
       _cmsg resp;
       // send MANUFACTURER_RESP
       MANUFACTURER_RESP_HEADER(&resp, mycmsg.ApplId, mycmsg.Messagenumber, CONTROLLER(line));
       MANUFACTURER_RESP_CONTROLLER(&resp) = MANUFACTURER_IND_CONTROLLER(&mycmsg);
       MANUFACTURER_RESP_MANUID(&resp) = MANUFACTURER_IND_MANUID(&mycmsg);
       MANUFACTURER_RESP_MANUDATA(&resp) = MANUFACTURER_IND_MANUDATA(&mycmsg);
       this->SendCapiMessage(line, &resp);
}*/
    else if (IS_RESET_B3_IND(&mycmsg)) {
        /* Network Control Connection Identifier */
        int ncci = RESET_B3_IND_NCCI(&mycmsg);
        unsigned char*ncpi = RESET_B3_IND_NCPI(&mycmsg);
        log->logf("<verbose> [line %d] got RESET_B3_IND: %x %x", line,ncci,ncpi);
        if(ncpi && *ncpi) {
            printf("ncpi length: %d\n", *ncpi);
            int t;
            for(t=1;t<*ncpi;t++) {
                printf("%02x ", ncpi[t]);
            }
        }
        _cmsg CMSGresp;
        RESET_B3_RESP_HEADER(&CMSGresp, mycmsg.ApplId, mycmsg.Messagenumber, CONTROLLER(line));
        RESET_B3_RESP_NCCI(&CMSGresp) = RESET_B3_RESP_NCCI(&mycmsg);
        this->SendCapiMessage(line, &CMSGresp);
    } else if (IS_INFO_IND(&mycmsg)) {
        /* we get these sometimes when we answer a call.
        * capi: is there something you want to tell us?
	*/
	
        int infonumber = INFO_IND_INFONUMBER(&mycmsg);
        log->logf("<verbose> [line %d] INFO_IND: infonumber:%02x", line,infonumber);
        if((infonumber&0xc000)==0) // non-supplementary information element
        {
            //we could now decipher this, according to ETS 300 102-1 Q.931.
            //we're just not in the mood.
        }

	/* These INFO_IND packets also fly by when some other application is answering
	   calls- be careful in doing too much here. */
        _cmsg CMSGresp;
        // send INFO_RESP
        INFO_RESP_HEADER(&CMSGresp, mycmsg.ApplId, mycmsg.Messagenumber, CONTROLLER(line));
        INFO_RESP_PLCI(&CMSGresp) = INFO_IND_PLCI(&mycmsg);
        this->SendCapiMessage(line, &CMSGresp);
    } else {
        log->logf("<error> [line %d] (MessageLoop) unknown message command=0x%02x subcommand=0x%02x", line, mycmsg.Command, mycmsg.Subcommand);
    }
    return 0;
}

/* fill the ncpi memory structure with correct values for x31 connections
 * using the specified datex destination address
 * @param ncpi pointer to the memory block used for the structure
 */
void CCapiHandler2::fillNCPIStruct (int line, unsigned char* ncpi) {
	/*        unsigned char ncpi[] = { 0x18, 0x00, 0x00, 0x00, 0xbb, 0x35, 0x69, 0x00, 0x11, 0x11, 0x04,
                                 0x59, 0x99, 0x99, 0x99, 0x98, 0x00, 0x01, 0x00, 0x00, 0x00,
                                 0x47, 0x49, 0x43, 0x43 };*/
    int addrLen = strlen(channeltab[line].other_datexp);
    int pos = 5;
    ncpi[1] = 0x00;
    ncpi[2] = 0x00;
    ncpi[3] = 0x00;
    if (addrLen > 15) {
    	ncpi[4] = 0x0f;
    } else {
    	ncpi[4] = addrLen;
    }
    for (int i = 0; i < addrLen; i++) {
    	ncpi[pos] = ((channeltab[line].other_datexp[i++] - '0') << 4) & 0xF0;
    	if (i < addrLen) {
    	    ncpi[pos] |= (channeltab[line].other_datexp[i] - '0') & 0x0F;
    	}
    	pos++;
    }
    ncpi[pos] = 0x00; // facility length
    ncpi[++pos] = 0x01; // user data
    ncpi[++pos] = 0x00;
    ncpi[++pos] = 0x00;
    ncpi[++pos] = 0x00;
    ncpi[++pos] = 0x47;
    ncpi[++pos] = 0x49;
    ncpi[++pos] = 0x43;
    ncpi[++pos] = 0x43;
    ncpi[0] = pos; // ncpi length
}

/* in principle, this routine is a finite state automon.
 * We've got some states, which are indictated by the state-variable
 * and some terminating states (which do a return)
 */
ISDNEvent CCapiHandler2::listen (int line, float&timeout) {
    log->logf("<verbose> [line %d] entering Message Loop...", line);
    //    unsigned int mexchError;   // MESSAGE_EXCHANGE_ERROR
    _cmsg mycmsg;
    // all purpose buffer

    int handletimeout =0;
    if (timeout>0)
        handletimeout =1;

    while (true) {
        if(channeltab[line].msgbufsize == 0) //buffer is empty
            mycmsg = getCapiMessage(line, timeout);
        else {
            mycmsg = channeltab[line].msgbuf[channeltab[line].msgbufreadpos];
            channeltab[line].msgbufreadpos++;
            channeltab[line].msgbufreadpos %= MSGBUFSIZE;
            channeltab[line].msgbufsize--;
        }
        if(timeout<=0 && handletimeout) {
            break; // timeout
        }
        else {
            ISDNEvent e;
            if(this->handleMessage(line, mycmsg, &e))
                return e;
        }
    }

    ISDNEvent e;
    e.type = TIMEOUT;
    e.line = line;
    e.msisdn = 0;
    e.reason = 0;
    e.jobid = -1;
    e.tone = -1;
    return e;
}

/* tries to forward the given line to the given msisdn. This still
 * doesn't work. We always get a "network did not answer in time"
 * message.
 */
int CCapiHandler2::forward (int line, char*msisdn) {
    /* this won't work with the windows driver of the
     * fritz card, which supports only supplementary services 1-4.
     */
    printf("STATE(line):%d state_idle:%d\n", STATE(line), state_idle);
    if(STATE(line) == state_alerted) // not connected
    {
        _cmsg req;
        FACILITY_REQ_HEADER(&req, APPL_ID(line), (unsigned short)this, CONTROLLER(line));
        FACILITY_REQ_PLCI(&req) = channeltab[line].plci;
        FACILITY_REQ_NCCI(&req) = channeltab[line].ncci;
        FACILITY_REQ_FACILITYSELECTOR(&req) = (unsigned short)3; /* 3: supplementary */
        // CF word 0x9 =CF
        //    dword handle
        //    word 0x0000 (unconditional)
        //    word basic service (ETS 300 196 (D.5))
        //    struct server user number (facility party number)
        //    struct forwarded to number (facility party number)
        //    struct forwarded to subaddress (called party subaddress)
        // CD word 0xD =CD
        //    word presentation allowed (0=no, 1=yes)
        //    struct deflected to number
        //    struct deflected to subaddress
        //
        //                         len CF     len handle   cond.    service  usernr            forward to         subaddress
        //unsigned char cBuffer[] = {20, 0x9,0, 17, 1,2,3,4, 0,0,     17,1,     3,0x81,0x33,0x33, 3,0x80,0x32,0x39,  0};
        //                        len CD     len pres.  number
        unsigned char cBuffer[] = {10, 0xd,0,  7, 1,0,   3,0x80,0x22,0x21,  0}; //CD - works ok
        //unsigned char cBuffer[] = { 3, 0x0,0, 0}; //INFO - works ok
        FACILITY_REQ_FACILITYREQUESTPARAMETER(&req) = cBuffer;
        this->SendCapiMessage(line, &req);
    }
    return 0;
}

/* puts a call on hold. The call won't be actually on hold
 * until a FACILITY_IND message is received
 */
int CCapiHandler2::hold(int line) {
    if(!channeltab[line].flag_hold) {
        _cmsg req;
        FACILITY_REQ_HEADER(&req, APPL_ID(line), (unsigned short)this, CONTROLLER(line));
        FACILITY_REQ_PLCI(&req) = channeltab[line].plci;
        FACILITY_REQ_NCCI(&req) = channeltab[line].ncci;
        FACILITY_REQ_FACILITYSELECTOR(&req) = (unsigned short)3; /* 3: supplementary */
        unsigned char cBuffer[] = {3, 0x2,0, 0}; // HOLD
        FACILITY_REQ_FACILITYREQUESTPARAMETER(&req) = cBuffer;
        this->SendCapiMessage(line, &req);
        this->waitForEvent(line, CAPI_FACILITY,CAPI_CONF, 1);
        channeltab[line].supplementaryjob = global_job_counter++;
        channeltab[line].state = state_holdreq;
        return channeltab[line].supplementaryjob;
    } else {
        log->logf("<error> [line %d] Can't put two lines on hold", line);
    }
    return 0;
}

/* tries to retrieve a call. The call won't be actually on retrieved
 * until a FACILITY_IND message is received.
 */
int CCapiHandler2::retrieve(int line) {
    if(channeltab[line].flag_hold) {
        _cmsg req;
        FACILITY_REQ_HEADER(&req, APPL_ID(line), (unsigned short)this, CONTROLLER(line));
        FACILITY_REQ_PLCI(&req) = channeltab[line].plci;
        FACILITY_REQ_NCCI(&req) = channeltab[line].ncci;
        FACILITY_REQ_FACILITYSELECTOR(&req) = (unsigned short)3; /* 3: supplementary */
        unsigned char cBuffer[] = {3, 0x3,0, 0}; // RETRIEVE
        FACILITY_REQ_FACILITYREQUESTPARAMETER(&req) = cBuffer;
        this->SendCapiMessage(line, &req);
        this->waitForEvent(line, CAPI_FACILITY,CAPI_CONF, 1);
        channeltab[line].supplementaryjob = global_job_counter++;
        channeltab[line].state = state_retrievereq;
        return channeltab[line].supplementaryjob;
    } else {
        log->logf("Can't do retrieve(). No call waiting.");
    }
    return 0;
}

/* tries to open up a 3pty with one held and one open call
 * at the moment, 3pty works, but may cause problems with
 * "unexpected" hangups, as we'll retrieve an hangup event
 * for both lines.
 */
int CCapiHandler2::conference(int line) {
    if(!channeltab[line].flag_hold) {
        log->logf("There's no line on hold. Can't do three party conference.");
        return 0;
    }
    if(channeltab[line].state != state_connected) {
        log->logf("There's no connection open. Can't do three party conference.");
        return 0;
    }

    _cmsg req;
    FACILITY_REQ_HEADER(&req, APPL_ID(line), (unsigned short)this, CONTROLLER(line));
    FACILITY_REQ_PLCI(&req) = channeltab[line].plci;
    FACILITY_REQ_NCCI(&req) = channeltab[line].ncci;
    FACILITY_REQ_FACILITYSELECTOR(&req) = (unsigned short)3; /* 3: supplementary */
    //                         len ID     len Call
    //unsigned char cBuffer[] = {7,  0x7,0, 4,  1,2,1,0}; // 3PTY
    unsigned char cBuffer[] = {7,  0x7,0, 4,  1,1,1,0}; // 3PTY
    // insert the plci of the line on hold into the struct
    *(int*)&cBuffer[4]=channeltab[line].flag_hold;
    FACILITY_REQ_FACILITYREQUESTPARAMETER(&req) = cBuffer;
    this->SendCapiMessage(line, &req);
    this->waitForEvent(line, CAPI_FACILITY,CAPI_CONF, 1);
    channeltab[line].supplementaryjob = global_job_counter++;
    channeltab[line].state = state_3ptyreq;
    return channeltab[line].supplementaryjob;
}

/* tries to perform ECT (explicit call transfer)
 * suffers from the same problem as CF
 */
int CCapiHandler2::ect(int line) {
    if(!channeltab[line].flag_hold) {
        log->logf("There's no line on hold. Can't do ect.");
        return 0;
    }
    if(channeltab[line].state != state_connected) {
        log->logf("There's no connection open. Can't do ect.");
        return 0;
    }
    _cmsg msg, req, resp;
    DISCONNECT_B3_REQ_HEADER(&req, APPL_ID(line), 1234, CONTROLLER(line));
    DISCONNECT_B3_REQ_NCCI(&req) = channeltab[line].ncci;
    DISCONNECT_B3_REQ_NCPI(&req) = NULL;
    SendCapiMessage(line, &req);

    FACILITY_REQ_HEADER(&req, APPL_ID(line), (unsigned short)this, CONTROLLER(line));
    FACILITY_REQ_PLCI(&req) = channeltab[line].plci;
    FACILITY_REQ_NCCI(&req) = channeltab[line].ncci;
    FACILITY_REQ_FACILITYSELECTOR(&req) = (unsigned short)3; /* 3: supplementary */
    //                         len ID     len Call
    unsigned char cBuffer[] = {7,  0x6,0, 4,  1,1,1,0}; // 3PTY
    // insert the plci of the line on hold into the struct
    *(int*)&cBuffer[4]=channeltab[line].flag_hold;
    FACILITY_REQ_FACILITYREQUESTPARAMETER(&req) = cBuffer;
    this->SendCapiMessage(line, &req);

    int num = 0;
    while(1) {
        float timeout = 60;
        msg = getCapiMessage(line, timeout);
        if(timeout <= 0)
            return 0;
        if(IS_DISCONNECT_B3_CONF(&msg)) {
            log->logf("<verbose> [line %d] disconnect b3 confirmed.", line);
        } else if(IS_DISCONNECT_CONF(&msg)) {
            log->logf("<verbose> [line %d] disconnect confirmed.", line);
        } else if(IS_FACILITY_IND(&msg)) {
            ISDNEvent e;
            if(!this->handleMessage(line, msg, &e))
                return 0; // don't reset line, as the disconnects won't happen
        }
        else if(IS_DISCONNECT_B3_IND(&msg)) {
            log->logf("<verbose> [line %d] b3 disconnected", line);
            DISCONNECT_B3_RESP_HEADER(&resp, APPL_ID(line), msg.Messagenumber, CONTROLLER(line));
            DISCONNECT_B3_RESP_NCCI(&resp) = DISCONNECT_B3_IND_NCCI(&msg);
            SendCapiMessage(line, &resp);
        } else if(IS_DISCONNECT_IND(&msg)) {
            log->logf("<notice> [line %d] disconnected", line);
            DISCONNECT_RESP_HEADER(&resp, APPL_ID(line), msg.Messagenumber, CONTROLLER(line));
            DISCONNECT_RESP_PLCI(&resp) = DISCONNECT_IND_PLCI(&msg);
            SendCapiMessage(line, &resp);
            num ++;
            if(num == 2) // both lines will be released
                break;
        }
    }
    resetLineToIdle(line);
    return 1;
}

/* Tries to answer a waiting call. This routine doesn't
 * take the line as parameter, but rather the INCOMING event
 * which is associated with the call to answer. The line
 * variable is extracted from that struct. Also, the internal
 * data of ISDNEvent (which, in this case, is the _cmsg of
 * the CONNECT_IND event) is used.
 */
int CCapiHandler2::answer (ISDNEvent *event) {
    int line = event->line;
    int ret = 0;
    _cmsg tmp;

    // pCMSG should be the CONNECT_IND message
    _cmsg*pCMSG=(_cmsg*)event->internal;
    tmp = *pCMSG;
    free(event->internal);
    pCMSG = &tmp;

    if (STATE(line) == state_alerted) {
        log->logf ("<notice> answering call on line %d\n", event->line);

        _cmsg mycmsg;

        CONNECT_RESP_HEADER(&mycmsg, APPL_ID(line), pCMSG->Messagenumber, CONTROLLER(line));
        CONNECT_RESP_PLCI(&mycmsg) = CONNECT_IND_PLCI(pCMSG);

        CONNECT_RESP_REJECT(&mycmsg) = 0;  // 0: accept call
        //CONNECT_RESP_REJECT(&mycmsg) = 2;  // 2: reject call, normal call clearing

        CONNECT_RESP_B1PROTOCOL(&mycmsg) = 1;
        CONNECT_RESP_B2PROTOCOL(&mycmsg) = 1;
        CONNECT_RESP_B3PROTOCOL(&mycmsg) = 0;

        ret = SendCapiMessage(line, &mycmsg);
        if (ret) {
            log->logf("<fatal> [line %d] answer:Could not send CONNECT_RESP message.", line);
            return 0;
        }

        // we assume that the responses come all and in exactly this
        // order. If that is not the case, we'd need something more
        // sophisticated.
        _cmsg resp, msg;

        // wait for CONNECT_ACTIVE_IND
        msg = waitForEvent(line, 0 /*CAPI_CONNECT_ACTIVE*/, CAPI_IND, 1);
        if(!IS_CONNECT_ACTIVE_IND(&msg)) {
            if(IS_DISCONNECT_IND(&msg)) {
                ISDNEvent e;
                handleMessage(line, msg, &e);
            } else
                insertIntoQueue(line, msg);
            log->logf("<notice> [line %d] waited for CONNECT_ACTIVE, but got %02x-%02x", line, msg.Command, msg.Subcommand);
            return 0;
        }

        // send CONNECT_ACTIVE_RESP
        CONNECT_ACTIVE_RESP_HEADER(&resp, msg.ApplId, msg.Messagenumber, CONTROLLER(line));
        CONNECT_ACTIVE_RESP_PLCI(&resp) = CONNECT_ACTIVE_IND_PLCI(&msg);
        ret = SendCapiMessage(line, &resp);
        if (ret) {
            log->logf("<fatal> [line %d] answer: Could not send CONNECT_ACTIVE_RESP message.", line);
            return 0;
        }

        // wait for CONNECT_B3_IND
        msg = waitForEvent(line, 0 /*CAPI_CONNECT_B3*/, CAPI_IND, 1);
        if(!IS_CONNECT_B3_IND(&msg)) {
            if(IS_DISCONNECT_IND(&msg)) {
                ISDNEvent e;
                handleMessage(line, msg, &e);
            } else
                insertIntoQueue(line, msg);
            log->logf("<notice> [line %d] waited for CONNECT_B3, but got %02x-%02x", line, msg.Command, msg.Subcommand);
            return 0;
        }

        // send CONNECT_B3_RESP
        CONNECT_B3_RESP_HEADER(&resp, msg.ApplId, msg.Messagenumber, CONTROLLER(line));
        CONNECT_B3_RESP_NCCI(&resp) = CONNECT_B3_IND_NCCI(&msg);
        CONNECT_B3_RESP_NCPI(&resp) = CONNECT_B3_IND_NCPI(&msg);
        ret = SendCapiMessage(line, &resp);
        if (ret) {
            log->logf("<fatal> [line %d] answer:Could not send CONNECT_B3_RESP message.", line);
            return 0;
        }

        // wait for CONNECT_B3_ACTIVE_IND
        msg = waitForEvent(line, 0 /*CAPI_CONNECT_B3_ACTIVE*/, CAPI_IND, 1);
        if(!IS_CONNECT_B3_ACTIVE_IND(&msg)) {
            if(IS_DISCONNECT_IND(&msg)) {
                ISDNEvent e;
                handleMessage(line, msg, &e);
            } else if(IS_DISCONNECT_B3_IND(&msg)) {
                ISDNEvent e;
                handleMessage(line, msg, &e);
                /* we're now going into the messageloop for some seconds, because
                 * there may be some disconnect_ind about to come up. we don't
                 * want it to pop up anywhere later, when it will maybe confuse
                 * some other routine.
                 */
                float timeout = 2;
                this->listen(line,timeout);
            } else {
                insertIntoQueue(line, msg);
            }
            log->logf("<notice> [line %d] waited for CONNECT_B3_ACTIVE, but got %02x-%02x", line, msg.Command, msg.Subcommand);
            return 0;
        }
        channeltab[line].ncci = CONNECT_B3_ACTIVE_IND_NCCI(&msg);
        channeltab[line].plci = mycmsg.adr.adrPLCI;
        log->logf("<verbose> [line %d] Setting NCCI to %x", line,channeltab[line].ncci);
        log->logf("<verbose> [line %d] Setting PLCI to %x", line,channeltab[line].plci);

        // send CONNECT_B3_RESP
        CONNECT_B3_ACTIVE_RESP_HEADER(&resp, msg.ApplId, msg.Messagenumber, CONTROLLER(line));
        CONNECT_B3_ACTIVE_RESP_NCCI(&resp) = CONNECT_B3_ACTIVE_IND_NCCI(&msg);
        ret = SendCapiMessage(line, &resp);
        if (ret) {
            log->logf("<fatal> [line %d] answer:Could not send CONNECT_B3_ACTIVE_RESP message.", line);
            return 0;
        }

        // try to enable dial tones

        log->logf("<verbose> [line %d] enabling dial tones...", line);
        // send FACILITY_REQ
        FACILITY_REQ_HEADER(&resp, msg.ApplId, msg.Messagenumber, CONTROLLER(line));
        FACILITY_REQ_NCCI(&resp) = CONNECT_B3_ACTIVE_IND_NCCI(&msg);
        FACILITY_REQ_FACILITYSELECTOR(&resp) = (unsigned short)1; /* 1: DTMF */
        /*len  listen  msdigit mspause tosend  characters */
        unsigned char cBuffer[] = {8,  1,0,    40,0,   40,0,   0,      0,         0};
        FACILITY_REQ_FACILITYREQUESTPARAMETER(&resp) = cBuffer;
        ret = SendCapiMessage(line, &resp);
        if(ret) {
            log->logf("<notice> [line %d] answer:Could not send FACILITY_REQ message. Dialtones not supported?", line);
        }

        // wait for CAPI_FACILITY_CONF
        msg = waitForEvent(line, CAPI_FACILITY, CAPI_CONF, 1);

        if (FACILITY_CONF_FACILITYSELECTOR(&msg) == 1) {  // facilityselector = 1 means DTMF supported
            FLAG_DTMF(line) = 1;
        } else
            log->logf("<error> [line %d] unknown facility_conf: %d (dialtones not supported?)", line,
                      FACILITY_CONF_FACILITYSELECTOR(&mycmsg));

        STATE(line) = state_connected;
        return 1;
    } else {
        log->logf ("<error> state=%d but though answer was called\n", STATE(line));
    }
    return 0;
}

/* disconnects the given line. It will only disconnect fully connected lines.
 * Also, it will reset the line_t struct.
 */
void CCapiHandler2::hangup (int line) {
    if (STATE(line) != state_connected &&
            STATE(line) != state_holdreq &&
            STATE(line) != state_3ptyreq) {
        log->logf("<error> [line %d] hangup: can't hangup - no connection open", line);
        return;
    }
    // b3 req
    _cmsg msg, resp;
    DISCONNECT_B3_REQ_HEADER(&msg, APPL_ID(line), 1234, CONTROLLER(line));
    DISCONNECT_B3_REQ_NCCI(&msg) = channeltab[line].ncci;
    DISCONNECT_B3_REQ_NCPI(&msg) = NULL;
    SendCapiMessage(line, &msg);

    DISCONNECT_REQ_HEADER(&msg, APPL_ID(line), 1235, CONTROLLER(line));
    DISCONNECT_REQ_PLCI(&msg) = channeltab[line].plci;
    SendCapiMessage(line, &msg);

    while(1) {
        float timeout = 60;
        msg = getCapiMessage(line, timeout);
        if(timeout <= 0)
            break;
        if(IS_DISCONNECT_B3_CONF(&msg)) {
            log->logf("<verbose> [line %d] disconnect b3 confirmed.", line);
        } else if(IS_DISCONNECT_CONF(&msg)) {
            log->logf("<verbose> [line %d] disconnect confirmed.", line);
        } else if(IS_DISCONNECT_B3_IND(&msg)) {
            log->logf("<verbose> [line %d] b3 disconnected", line);
            DISCONNECT_B3_RESP_HEADER(&resp, APPL_ID(line), msg.Messagenumber, CONTROLLER(line));
            DISCONNECT_B3_RESP_NCCI(&resp) = DISCONNECT_B3_IND_NCCI(&msg);
            SendCapiMessage(line, &resp);
        } else if(IS_DISCONNECT_IND(&msg)) {
            log->logf("<notice> [line %d] disconnected", line);
            DISCONNECT_RESP_HEADER(&resp, APPL_ID(line), msg.Messagenumber, CONTROLLER(line));
            DISCONNECT_RESP_PLCI(&resp) = DISCONNECT_IND_PLCI(&msg);
            SendCapiMessage(line, &resp);
            break;
        }
    }

    resetLineToIdle(line);
}

/* convert an event into readable form. does allocate a
 * string, which you should free afterwards!
 */
char* CCapiHandler2::eventToString (ISDNEvent*event) {
    char buf[256];
    char*name;
    switch(event->type) {
    case TIMEOUT:
        name="TIMEOUT";
        break;
    case CAPI_EVENT_HANGUP:
        name="HANGUP";
        break;
    case CAPI_EVENT_INCOMING:
        name="INCOMING";
        break;
    case CAPI_EVENT_DONE:
        name="DONE";
        break;
    case CAPI_EVENT_DIALTONE:
        name="DIALTONE";
        break;
    case CAPI_EVENT_CONNECTED:
        name="CONNECTED";
        break;
    }
    int s=sprintf(buf, "EVENT [%d] name=%s msisdn=\"%s\"", event->line, name, event->msisdn);
    if (event->type == CAPI_EVENT_DIALTONE)
        sprintf(&buf[s], " tone='%c'", event->tone);
    if (event->type == CAPI_EVENT_DONE)
        sprintf(&buf[s], " job=%d", event->jobid);
    if (event->type == CAPI_EVENT_HANGUP && event->reason)
        sprintf(&buf[s], " reason=%d", event->reason);
    return strdup(buf);
}

