/* CapiHandler2.h
 *
header for CapiHandler2.cpp

Part of the caiviar package. 

Copyright (c) 2002 MobileX AG, http://www.mobilexag.de
Copyright (c) 2002 Matthias Kramm <kramm@quiss.org>

This file is distributed under the GPL, see file COPYING for details. */

#define MAXCHANNELS 128

#if !defined(AFX_CAPIHANDLER_H__1747CA67_0D60_11D5_810E_AEC6F7122744__INCLUDED_)
#define AFX_CAPIHANDLER_H__1747CA67_0D60_11D5_810E_AEC6F7122744__INCLUDED_

#define MSGBUFSIZE 64

#ifdef LINUX
# include <stdio.h>
# include <stdlib.h>
# include <pthread.h>
# include "Log.h"
# include "clinux.h"
# include "SimpleCapi.h"
#else
# if _MSC_VER > 1000
# pragma once
#endif // _MSC_VER > 1000

#ifdef CYGWIN
#include "cwin32.h"
#endif

#include <winsock2.h>
#include <stdlib.h>
#include <malloc.h>

#include "Log.h"

#include "capi20.h"

#include <windef.h>
#include <winbase.h>
#include "SimpleCapi.h"
#endif

#ifdef __linux__
typedef struct x31proto3
{
    unsigned char size;
    unsigned short lic __attribute__ ((packed));;
    unsigned short hic __attribute__ ((packed));;
    unsigned short ltc __attribute__ ((packed));;
    unsigned short htc __attribute__ ((packed));;
    unsigned short loc __attribute__ ((packed));;
    unsigned short hoc __attribute__ ((packed));;
    unsigned short modulomode __attribute__ ((packed));
    unsigned short windowsize __attribute__ ((packed));
} B3_PROTO_X31;
#else
#pragma pack (1)
typedef struct x31proto3
{
    unsigned char size;
    unsigned short lic;
    unsigned short hic;
    unsigned short ltc;
    unsigned short htc;
    unsigned short loc;
    unsigned short hoc;
    unsigned short modulomode;
    unsigned short windowsize;
} B3_PROTO_X31;
#pragma pack ()
#endif


/* these are the states of some finite state automon.
 * The comments are the other states that can be reached from the
 * actual state. The functions in parentheses are the ones which
 * can be called for the state transition in question
 */
enum state_t
{
    state_idle,  // alertreq, alerted(DONT_DO_ALERT=1), calling(call())
    state_calling,  // connected, idle
    state_alertreq, // alerted
    state_alerted, // connected(answer()), idle(answer() failed)
    state_connected, // disconnecting(hangup from other end), idle(hangup())
    state_disconnecting, // idle
    state_holdreq,
    state_retrievereq,
    state_3ptyreq
};


/* at the moment, this job structure is only used for play and
 * record jobs. call is also a job, but doesn't have a structure.
 * The job id for calls is always 0x0300+line
 */
struct job {
    char * data;
    int data_size;
    char * data_pos;
    int bytes_remaining;
    short unsigned int datahandle;
    int jobid;
} ;


struct Line_t {
    char active; // 1=in use, 0=idle

    // as far as known. set e.g. in handleMessage(), when a call comes in.
    unsigned int plci;
    int ncci;

    // the controller this line belongs to. notice that the line is
    // arbitrary, as the B-channels of controllers are usually not ordered.
    int controller;
    int controller_line;

    // the msisdn of the connected or calling instance
    char other_msisdn[20];

    // datexp address to be called
    char other_datexp[20];
    
    // my own msisdn. Not neccessarily the same for all lines.
    char*my_msisdn;

    // the job id of the calling job.
    int callingjob;

    int flag_listen; // are we listening for calls?
    int flag_dtmf; // are dtmf enabled/possible with the running connection?
    int flag_fax; // are we connected to a fax or a person?
    int flag_x31; // is this a x31 connection?
    int flag_hold; // is there another call waiting? (if yes flag_hold holds it's plci)
    int flag_3pty; // are we connected to 2 partys simultaneously?

    state_t state; // the state we are in. see definition of state_t

    unsigned short Appl_Id; // the appl_id of this line. must be used in all send_message calls

    job playjob; // max one per line
    job recordjob; // likewise
    int supplementaryjob;

    _cmsg*incoming;

    /* these variables are a message queue for
     * messages that come in while we are waiting for 
     * something else, e.g. in waitForEvent.
     */
    _cmsg msgbuf[MSGBUFSIZE];
    int msgbufreadpos;
    int msgbufwritepos;
    int msgbufsize;
};



class CCapiHandler2  
{
    friend SimpleCapi;
public:
	CCapiHandler2(Logger*log);
	virtual ~CCapiHandler2();
	
	int open (char*msisdn, char*controller);
	void close (int line);
	void acceptCalls (int line);
	void setDatexAddress (int line, char* address);
	int call (int line, char* msisdn, int mode);
	int play (int line, void*mem, int len); 
	int record (int line, void*buf, int bufsize);
	int getRecordSize (int line);
	//int fax (int line, void*fax, int faxsize);
	void stop (int jobid);
	ISDNEvent listen (int line, float&timeout);
	int answer (ISDNEvent*event);
	int forward (int line, char*msisdn);
	void hangup (int line);
	char* eventToString (ISDNEvent*event);
	int checkLine(int line);

	int hold (int line);
	int retrieve (int line);
	int conference(int line);
	int ect(int line);
private:
	void resetLineToIdle(int line);
	int handleMessage(int line, _cmsg mycmsg, ISDNEvent*e);
	void insertIntoQueue(int line, _cmsg mymsg);
	_cmsg getCapiMessage(int line, float&timeout);
	_cmsg waitForEvent(int line,int event, int subevent, int store);
	void transferData(int line);
	int openLine(int line, char*msisdn);
	void fillNCPIStruct(int line, unsigned char* ncpi);

	int init();
	 unsigned int InitCAPI(void); //called from init

	int translatereason(int line, int reason);
        void showSupplementaryReason(int line, short int reason);

	unsigned int SendCapiMessage(int line, _cmsg* CMSG);
	unsigned int GetNumOfSupportedBChannels(long Controller);

	Logger*log;
	int maxchannels;
	int allcalls_hack;

	struct Line_t channeltab[MAXCHANNELS];

#if defined(LINUX)
   pthread_mutex_t capi_mutex;
#elif defined(WIN32)
   CRITICAL_SECTION capi_mutex;
#endif
};

#endif // !defined(AFX_CAPIHANDLER_H__1747CA67_0D60_11D5_810E_AEC6F7122744__INCLUDED_)
