/* Log.cpp

generic Logging class for file, screen and network logging.

Part of the caiviar package. 

Copyright (c) 2002 MobileX AG, http://www.mobilexag.de
Copyright (c) 2002 Matthias Kramm <kramm@quiss.org>

This file is distributed under the GPL, see file COPYING for details. */

#ifdef LINUX
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#define EnterCriticalSection pthread_mutex_lock
#define LeaveCriticalSection pthread_mutex_unlock
#define InitializeCriticalSection(a) 
#define DeleteCriticalSection(a) 
#else
#include <string.h>
#include <winsock2.h>
#include <stdlib.h>
#include <malloc.h>
#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000
#endif

#include "Log.h"

void Logger::initLog(char* pLogName, int fileloglevel, char* servAddr, char* logPort, int serverlevel, int screenlevel) {
    this->screenloglevel = screenlevel;
    this->fileloglevel = fileloglevel;
    this->serverloglevel = screenlevel;
    logFile = NULL;
    bLogToSock = false;

    logFile = fopen(pLogName, "a+");
    bLogToSock = (servAddr && logPort);
    if(bLogToSock)
        initlogSocket(servAddr, logPort);
    InitializeCriticalSection(&this->log_mutex);
}

void Logger::initlogSocket(char* servAddr, char* logPort) {
#ifdef LINUX
    bLogToSock = 0;
#else
// init winsock
    // check and prepare WinSock DLL
    WORD wVersionRequested = MAKEWORD( 2, 2 );
    WSADATA wsaData;
    if ( WSAStartup(wVersionRequested, &wsaData) != 0 ) {
        bLogToSock = false;
        return;
    }
    // Confirm that the WinSock DLL supports 2.2.
    // Note that if the DLL supports versions greater
    // than 2.2 in addition to 2.2, it will still return
    // 2.2 in wVersion since that is the version we
    // requested.

    if ( LOBYTE( wsaData.wVersion ) != 2 || HIBYTE( wsaData.wVersion ) != 2 ) {
        bLogToSock = false;
        return;
    }

    struct hostent *hp;
    hp = gethostbyname(servAddr);
    if (hp == NULL) // we don't know who this host is
    {
        bLogToSock = false;
        return;
    }

    // connect socket
    sockaddr_in SocketAddress;

    memset(&SocketAddress, 0, sizeof(SocketAddress));
    memcpy((char*)&SocketAddress.sin_addr, hp->h_addr, hp->h_length); // set address
    SocketAddress.sin_family = hp->h_addrtype;
    SocketAddress.sin_port = htons((u_short)atoi(logPort));

    logSocket = socket(hp->h_addrtype, SOCK_STREAM, 0);
    if (logSocket == INVALID_SOCKET) {
        bLogToSock = false;
        return;
    }

    // try to connect to the specified socket
    if ( connect(logSocket, (struct sockaddr*)&SocketAddress, sizeof (SocketAddress)) == SOCKET_ERROR) {
        bLogToSock = false;
        return;
    }
    bLogToSock = true;
#endif
}

void Logger::exitLog() {
    // close socket communication
    if(bLogToSock)
#ifdef LINUX
        close(logSocket);
#else
        closesocket(logSocket);
#endif
    // close file
    if(logFile != NULL)
        fclose(logFile);
    DeleteCriticalSection(&this->log_mutex);
}

Logger::~Logger() {
    exitLog();
}

static char * logimportance[]= {"Fatal","Error","Warning","Notice","Verbose","Debug"};
static int loglevels=6;
static char * logimportance2[]= {"       ","FATAL  ", "ERROR  ", "WARNING", 
				 "NOTICE ","VERBOSE", "DEBUG  "};
void Logger::log(char* logString) {
    EnterCriticalSection(&this->log_mutex);
    char timebuffer[32];
    char* logBuffer;
    logBuffer = (char*)malloc (strlen(logString) + 24 + 15);
#ifdef LINUX
    {
        /*time_t t = time(0);
        tm*t2 = localtime(t);
        strftime(dbuffer, 8, "%m %d", t2);
        strftime(tbuffer, 8, "%m %d", t2);
        dbuffer[0]=0; //FIXME
        tbuffer[0]=0;*/
        time_t t = time(0);
        char* a = ctime(&t);
        int l = strlen(a);
        while(a[l-1] == 13 || a[l-1] == 10)
            l--;
        a[l]=0;
        sprintf(timebuffer, "%s", a);
    }
#else
char dbuffer[9];
    char tbuffer[9];
    _strdate( dbuffer );
    _strtime( tbuffer );
    sprintf(timebuffer, "%s - %s",dbuffer,tbuffer);
#endif

    // search for <level> field
    int level=-1;
    char*lt=strchr(logString, '<');
    char*gt=strchr(logString, '>');
    if(lt && gt && lt<gt) {
        int t;
        for(t=0;t<loglevels;t++) {
#ifdef LINUX
            if(!strncasecmp(lt+1,logimportance[t],strlen(logimportance[t])))
#else
            if(!strnicmp(lt+1,logimportance[t],strlen(logimportance[t])))
#endif
            {
                logString = gt+1;
                while(logString[0]==' ')
                    logString ++;
                level = t;
                break;
            }
        }
    }

    sprintf(logBuffer, "%s: %s %s", timebuffer, logimportance2[level + 1],logString);

    // we always do exactly one newline.

    int l=strlen(logBuffer)-1;
    while((logBuffer[l]==13 || logBuffer[l]==10) && l>=0) {
        logBuffer[l]=0;
        l--;
    }

    if (level <= screenloglevel) {
        printf("%s\n",logBuffer);
        fflush(stdout);
    }

    if (level <= fileloglevel) {
        if (logFile != NULL) {
            fprintf(logFile, "%s\n", logBuffer);
            fflush(logFile);
        }
    }

    if (level <= serverloglevel) {
        if (bLogToSock) {
            // send data
#ifdef LINUX
            write(logSocket, logBuffer, strlen(logBuffer));
#else
send(logSocket, logBuffer, strlen(logBuffer), 0);
#endif
        }
    }
    free (logBuffer);
    LeaveCriticalSection(&this->log_mutex);
}

void Logger::logf(const char* pszFormat, ...) {
    char buf[1024];
    sprintf(buf, "[%s]: ", logimportance[2]);
    va_list arglist;
    va_start(arglist, pszFormat);
    vsprintf(&buf[strlen(buf)], pszFormat, arglist);
    va_end(arglist);
    strcat(buf, "\n");
    log(buf);
}

