/* CapiHandler.cpp
 *
implementation of the CCapiHandler2 class.

Part of the caiviar package. 

Copyright (c) 2002 MobileX AG, http:
Copyright (c) 2002 Matthias Kramm <kramm@quiss.org>

This file is distributed under the GPL, see file COPYING for details. */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <memory.h>
#include <string.h>
#include <stdarg.h>
#include "wav2law.h"

#define min(a,b) ((a)>(b)?(a):(b))

int align2(int);

struct chunk {
    unsigned long int ckID;
    unsigned long int ckSize;
};

unsigned long int fourcc(char*);
chunk* basic_chunk(unsigned long int, unsigned long, void*);
chunk* ICMT_chunk(char*);

unsigned char PCM16_TO_ALAW(short int PCM_Value);
short int ALAW_TO_PCM16(unsigned char ALAW_Value);

template
<class T>
inline T endian(T x) {
    short bot=(short) 0xAABB;
    unsigned char h;
    unsigned char *b;
    int len=sizeof(x);

    if(*(unsigned char*)&bot==0xAA) 
    {
        b=(unsigned char*)&x;
        for(int i=0;i<len/2;i++) {
            h=b[i];
            b[i]=b[len-i-1];
            b[len-i-1]=h;
        }
    }
    return x;
}

struct fmt {
    unsigned short 	formatTag;
    unsigned short 	nChannels;
    unsigned long	nSamplesPerSec;
    unsigned long	nAvgBytesPerSec;
    unsigned short	nBlockAlign;
    unsigned short	nBitsPerSample;
};

chunk* fmt_chunk(unsigned short formatTag, unsigned short nChannels, unsigned long nSamplesPerSec,
                 unsigned long nAvgBytesPerSec, unsigned short nBlockAlign, unsigned short nBitsPerSample) {
    fmt f;
    f.formatTag = endian(formatTag);
    f.nChannels = endian(nChannels);
    f.nSamplesPerSec = endian(nSamplesPerSec);
    f.nAvgBytesPerSec = endian(nAvgBytesPerSec);
    f.nBlockAlign = endian(nBlockAlign);
    f.nBitsPerSample = endian(nBitsPerSample);

    return basic_chunk(fourcc("fmt "), sizeof(f), &f);
} 








#define SCALE_P16_TO_ALAW(x)  ((x)/8)
#define SCALE_ALAW_TO_P16(x)  ((x)*8)
#define FUDGE  2      
#define Offset_Mask		0x0F
#define DREHEN			0x55
#define Segment_Mask	0x70

unsigned char PCM16_TO_ALAW(short int PCM_Value) {
    char PlusMinus;
    unsigned short int PCM16BIT;
    unsigned char ALAW;
    unsigned char Carry;

    if (PCM_Value >=0)							
        PlusMinus = 1;
    else
        PlusMinus = -1;
    PCM16BIT=PCM_Value * PlusMinus;				
    if (PCM16BIT < 0x0800)						
    {
        if (PCM16BIT < 0x0200)					
        {
            if (PCM16BIT < 0x0100)				
            {
                if (PCM16BIT & 0x000c)			
                    Carry=1;
                else
                    Carry=0;
                PCM16BIT&=0x00F0;				
                PCM16BIT>>=3;					
                if (PlusMinus==1)
                    ALAW=0x80+PCM16BIT+Carry; 	
                else
                    ALAW=(unsigned char)PCM16BIT;
            } else								
            {
                if (PCM16BIT & 0x001c)			
                    Carry=1;
                else
                    Carry=0;
                PCM16BIT&=0x00E0;
                PCM16BIT>>=4;
                if (PlusMinus==1)
                    ALAW=0xA0+PCM16BIT+Carry; 	
                else
                    ALAW=0x20+PCM16BIT;
            }
        } else									
        {
            if (PCM16BIT <0x0400)				
            {
                if (PCM16BIT & 0x003c)			
                    Carry=1;
                else
                    Carry=0;
                PCM16BIT&=0x01C0;
                PCM16BIT>>=5;
                if (PlusMinus==1)
                    ALAW=0xB0+PCM16BIT+Carry; 	
                else
                    ALAW=0x30+PCM16BIT;
            } else								
            {
                if (PCM16BIT & 0x007c)			
                    Carry=1;
                else
                    Carry=0;
                PCM16BIT&=0x0380;
                PCM16BIT>>=6;
                if (PlusMinus==1)
                    ALAW=0xC0+PCM16BIT+Carry; 	
                else
                    ALAW=0x40+PCM16BIT;
            }
        }
    } else {
        if (PCM16BIT < 0x2000)					
        {
            if (PCM16BIT < 0x1000)				
            {
                if (PCM16BIT & 0x00Fc)			
                    Carry=1;
                else
                    Carry=0;
                PCM16BIT&=0x0700;
                PCM16BIT>>=7;
                if (PlusMinus==1)
                    ALAW=0xD0+PCM16BIT+Carry; 	
                else
                    ALAW=0x50+PCM16BIT;
            } else								
            {
                if (PCM16BIT & 0x01Fc)			
                    Carry=1;
                else
                    Carry=0;
                PCM16BIT&=0x0E00;
                PCM16BIT>>=8;
                if (PlusMinus==1)
                    ALAW=0xE0+PCM16BIT+Carry; 	
                else
                    ALAW=0x60+PCM16BIT;

            }
        } else									
        {
            if (PCM16BIT < 0x4000)				
            {
                if (PCM16BIT & 0x03Fc)			
                    Carry=1;
                else
                    Carry=0;
                PCM16BIT&=0x1C00;
                PCM16BIT>>=9;
                if (PlusMinus==1)
                    ALAW=0xF0+PCM16BIT+Carry; 	
                else
                    ALAW=0x70+PCM16BIT;

            } else								
                if (PlusMinus==1)
                    ALAW=0xFF; 	
                else
                    ALAW=0x7F;

        }
    }
    return	(ALAW ^ DREHEN);					
}

short int ALAW_TO_PCM16(unsigned char ALAW_Value) {
    unsigned char	Segment;
    unsigned char	Offset;
    short int	PCM_Value;

    ALAW_Value^=DREHEN;
    Segment=ALAW_Value & Segment_Mask;
    Segment>>=4;
    Offset=ALAW_Value & Offset_Mask;
    if (Segment < 0x04) {
        if (Segment < 0x02) {
            if (Segment < 0x01)
                PCM_Value=(0x0001+(Offset<<1))<<2;
            else
                PCM_Value=(0x0021+(Offset<<1))<<2;
        } else {
            if (Segment < 0x03)
                PCM_Value=(0x0042+(Offset<<2))<<2;
            else
                PCM_Value=(0x0084+(Offset<<3))<<2;
        }
    } else {
        if (Segment < 0x06) {
            if (Segment < 0x05)
                PCM_Value=(0x0108+(Offset<<4))<<2;
            else
                PCM_Value=(0x0210+(Offset<<5))<<2;
        } else {
            if (Segment < 0x07)
                PCM_Value=(0x0420+(Offset<<6))<<2;
            else
                PCM_Value=(0x0840+(Offset<<7))<<2;
        }
    }
    if (ALAW_Value<0x80)	
        PCM_Value*=-1;
    return PCM_Value;
}

class WaveFile {
private:
    int filesize;
    int riffpos, riffsize;
    int listpos, listsize;
    int icmtpos, icmtsize;
    int fmtpos,  fmtsize;
    int datasize;
    long datapos;
    chunk cbuf;

protected:
    char *comment;
    unsigned char *data;
    int  datalen;

    FILE* ipFile;
    char* ipFilename;

    char isinmemory;
    unsigned char* mem;
    unsigned char* mempos;
    int memlen;

    int myfseek(FILE*fi, int pos, int fj);
    int myftell(FILE*fi);
    int myfread(void*data, int bsize, int bnum, FILE*fi);
public:
    fmt  format;

    WaveFile(char*);
    WaveFile(unsigned char*, int len);
    virtual ~WaveFile();

    void print_info();

    int read_header(void);
    int readHeaderAndFormat(void);
    int read(void);
    int readDataBlock(char* aByteBuffer, int aNoOfBytes);
    int write(char* =0);

    int setcomment(char*);
    char * getcomment(void);

    int setdata(unsigned char*,int);
    unsigned char * getdata(void);

    int samplelen(void);
    int filesamplelen(void);

    int getSamplesPerSec() {
        return format.nSamplesPerSec;
    }

    int getchunk(FILE*, int, int, unsigned long int, int&);

    unsigned long int getformtype(FILE*, int);
};

void WaveFile::print_info() {
    printf("formatTag: %x\n", format.formatTag);
    printf("nChannels: %d\n", format.nChannels);
    printf("SamplesPerSec: %d\n", format.nSamplesPerSec);
    printf("avgbytespersec: %d\n", format.nAvgBytesPerSec);
    printf("blockAlign: %d\n", format.nBlockAlign);
    printf("bitspersample: %d\n", format.nBitsPerSample);
    printf("FileSize: %d\n", filesize);
}

int WaveFile::myfseek (FILE*fi, int pos, int fj) {
    if(!isinmemory)
        return fseek(fi, pos, fj);
    else {
        if(fj == SEEK_SET) {
            mempos = mem + pos;
        }
        if(fj == SEEK_END) {
            mempos = &mem [memlen + pos];
        }
        if(fj == SEEK_CUR) {
            mempos += pos;
        }
        return 0;
    }
}

int WaveFile::myftell (FILE*fi) {
    if(isinmemory) {
        return mempos - mem;
    } else
        return ftell(fi);
}

int WaveFile::myfread(void*data, int bsize, int bnum, FILE*fi) {
    if(isinmemory) {
        int t;
        for(t=0;t<bnum;t++) {
            if(mempos + bsize > mem + memlen)
                break;
            memcpy(data,mempos,bsize);
            mempos += bsize;
            data = &(((char*)data)[bsize]);
        }
        
        return t;
    } else
        return fread(data,bsize,bnum,fi);
}

int WaveFile::read_header(void) {
    
    setcomment("");
    delete data;
    data=0;

    if(NULL==ipFile && !isinmemory) {
        
        return __LINE__;
    }

    

    myfseek(ipFile,0,SEEK_END);
    filesize=myftell(ipFile);
    myfseek(ipFile,0,SEEK_SET);
    

    
    riffpos=getchunk(ipFile,
                     0,
                     filesize,
                     fourcc("RIFF"),
                     riffsize);
    if(riffpos<=0) 
    {
        
        return __LINE__;
    }

    
    if(getformtype(ipFile, riffpos)!=fourcc("WAVE")) {
        
        return __LINE__;
    }

    
    listpos=getchunk(ipFile,
                     riffpos+sizeof(unsigned long int),
                     riffsize-sizeof(unsigned long int),
                     fourcc("LIST"),
                     listsize);
    if(listpos<0) 
    {
        
        return __LINE__;
    } else
        if(listpos==0) 
        {
            setcomment("");
        } else 
            
            if(getformtype(ipFile, listpos)!=fourcc("INFO")) {
                setcomment("");
            } 
            else {	
                icmtpos=getchunk(ipFile,
                                 listpos+sizeof(unsigned long int),
                                 listsize-sizeof(unsigned long int),
                                 fourcc("ICMT"),
                                 icmtsize);
                if(icmtpos<0) 
                {
                    
                    return __LINE__;
                } 
                else
                    if(icmtpos==0) 
                    {
                        setcomment("");
                    } 
                    else {
                        char *buf=new char[icmtsize+1];
                        assert(buf);

                        myfread(buf, 1, icmtsize, ipFile);
                        buf[icmtsize]=0;
                        setcomment(buf);
                        delete buf;
                    } 
            } 

    

    fmtpos=getchunk(ipFile,
                    riffpos+sizeof(unsigned long int),
                    riffsize-sizeof(unsigned long int),
                    fourcc("fmt "),
                    fmtsize);
    if(fmtpos<=0 /*|| fmtsize!=sizeof(fmt)*/) 
    {
        
        return __LINE__;
    } 


    return 0;
} 


int WaveFile::read(void) {
    int errc=read_header();
    if(errc)
        return errc;


    memset(&format,0,sizeof(format));
    myfseek(ipFile, fmtpos,SEEK_SET);
    myfread(&format, 1, min(fmtsize,sizeof(format)), ipFile);

    format.formatTag	=endian(format.formatTag);
    format.nChannels	=endian(format.nChannels);
    format.nSamplesPerSec	=endian(format.nSamplesPerSec);
    format.nAvgBytesPerSec	=endian(format.nAvgBytesPerSec);
    format.nBlockAlign	=endian(format.nBlockAlign);
    format.nBitsPerSample	=endian(format.nBitsPerSample);

    
    
    
    fmt format08={1,1, 8000,16000,2,16};
    fmt format16={1,1,16000,32000,2,16};
    if(memcmp(&format,&format08,sizeof(fmt)) &&
            memcmp(&format,&format16,sizeof(fmt))) {
        return __LINE__;
    } 


    
    datapos=getchunk(ipFile,
                     riffpos+sizeof(unsigned long int),
                     riffsize-sizeof(unsigned long int),
                     fourcc("data"),
                     datasize);
    if(datapos<=0) 
    {
        return __LINE__;
    } 
    else {	
        myfseek(ipFile, datapos,SEEK_SET);
        unsigned char *buf=new unsigned char[datasize];
        
        assert(buf);

        myfread(buf, 1, datasize, ipFile);

        
        short *p=(short*)buf;
        for(int i=0;i<datasize/(int)sizeof(short); i++) {
            *p=endian(*p);
            p++;
        } 

        setdata(buf, datasize);
        delete buf;

    } 

    return 0;
} 

int WaveFile::readHeaderAndFormat(void) {
    int errc=read_header();
    if(errc) {
        
        return errc;
    }

    memset(&format,0,sizeof(format));
    myfseek(ipFile, fmtpos,SEEK_SET);
    myfread(&format, 1, min(fmtsize,sizeof(format)), ipFile);

    format.formatTag	=endian(format.formatTag);
    format.nChannels	=endian(format.nChannels);
    format.nSamplesPerSec	=endian(format.nSamplesPerSec);
    format.nAvgBytesPerSec	=endian(format.nAvgBytesPerSec);
    format.nBlockAlign	=endian(format.nBlockAlign);
    format.nBitsPerSample	=endian(format.nBitsPerSample);

    
    datapos=getchunk(ipFile,
                     riffpos+sizeof(unsigned long int),
                     riffsize-sizeof(unsigned long int),
                     fourcc("data"),
                     datasize);

    if(datapos<=0) 
    {
        return __LINE__;
    } 
    else {
        myfseek(ipFile, datapos,SEEK_SET);
    } 

    return 0;
} 

int WaveFile::readDataBlock(char* aByteBuffer, int aNoOfBytes) {
    int aReturn = myfseek(ipFile, datapos,SEEK_SET);
    if(aReturn) {
        return -1;
    }

    assert(aByteBuffer);

    int aNumberRead = 0;

    
    if(1) {
        if(!isinmemory)
            aNumberRead = (int) myfread(aByteBuffer,1,aNoOfBytes,ipFile);
        else { 
            int size = aNoOfBytes;
            if(mempos + size > mem + memlen)
                size = (mem + memlen) - mempos;
            memcpy(aByteBuffer, mempos, size);
            mempos += size;
            aNumberRead = size;
        }
    }
    datapos=myftell(ipFile);
    if(datapos<0) 
    {
        return -1;
    }
    return aNumberRead;
}

int WaveFile::setdata(unsigned char * ndata, int ndatalen) {
    delete data;
    data=0;
    datalen=0;

    if(ndata==0 || ndatalen<=0) {
        return  0;
    }
    data=new unsigned char[ndatalen];
    assert(data);

    memcpy(data, ndata, ndatalen);
    datalen=ndatalen;

    return 0;
} 


WaveFile::WaveFile(char *nfilename) :
        comment(""),
        data(0),
        datalen(0),
        datapos(0),
        ipFile(0),
        ipFilename(0),
isinmemory(0) {
    assert(nfilename);
    assert(strlen(nfilename));
    ipFilename = new char[strlen(nfilename)+1];
    assert(ipFilename);
    strcpy(ipFilename,nfilename);
    ipFile=fopen(ipFilename, "rb");

    fmt default_fmt={1,1,8000,16000,2,16};
    format=default_fmt;

} 

WaveFile::WaveFile(unsigned char *mem, int len) :
        comment(""),
        data(0),
        datalen(0),
        datapos(0),
        ipFile(0),
        ipFilename(0),
isinmemory(1) {
    this->mem = mem;
    this->memlen = len;
    this->mempos = mem;
    fmt default_fmt={1,1,8000,16000,2,16};
    format=default_fmt;

} 


WaveFile::~WaveFile() {
    if(ipFile) {
        fclose(ipFile);
        ipFile = 0;
    }
    delete data;
    delete [] ipFilename;
} 


char *WaveFile::getcomment(void) {
    return comment;
} 


int WaveFile::setcomment(char* ncomment) {
    if(comment)
        comment=ncomment;
    else
        comment="";

    return 0;
} 


unsigned char* WaveFile::getdata(void) {
    return data;
} 

int WaveFile::samplelen(void) {
    return datalen/(format.nBlockAlign);
} 

int WaveFile::filesamplelen(void) {
    return datasize/(format.nBlockAlign);
} 




int align2(int n) {
    return n+(n%2);
} 

unsigned char* inew(int len) {
    unsigned char *help=new unsigned char[len];
    assert(help);
    memset(help,0,len);
    return help;
} 


unsigned long int fourcc(char* s) {
    unsigned long int fcc=0x20202020;
    if(s==0 || *s==0)
        return fcc;

    strncpy((char*)&fcc,s,sizeof(fcc));
    if(strlen(s)<sizeof(fcc))
        *((char*)&fcc+strlen(s))=0x20;
    return fcc;
} 


chunk* basic_chunk(unsigned long int ckID, unsigned long ckSize, void *p) {
    if(ckSize)
        assert(p);
    
    
    chunk *ergch=(chunk*) inew(sizeof(chunk)+align2(ckSize));
    ergch->ckID=ckID;
    ergch->ckSize=endian(ckSize);
    if(ckSize)
        memcpy(ergch+1, p, ckSize);

    return ergch;
} 


chunk* ICMT_chunk(char *comment) {
    assert(comment);

    return basic_chunk(fourcc("ICMT"), strlen(comment), comment);
} 


int WaveFile::getchunk(	FILE* file, 	
                        int startpos,	
                        int cckSize,	
                        unsigned long int sckID,	
                        int &sckSize)	
{
    assert(file || isinmemory);

    chunk chunkbuf;
    int remain=align2(cckSize);

    if(0!=myfseek(file, startpos, SEEK_SET)) {
        return -1;
    }

    while(remain > 0) {
        if(sizeof(chunk)!=myfread(&chunkbuf, 1, sizeof(chunk), file)) {
            return -1;
        }
        remain-=sizeof(chunk);

        

        if(chunkbuf.ckID==sckID) {
            sckSize=endian(chunkbuf.ckSize);
            return myftell(file);
        } 

        if(0!=myfseek(file, align2(endian(chunkbuf.ckSize)),SEEK_CUR)) {
            return -1;
        }
        remain-=align2(endian(chunkbuf.ckSize));
    } 
    

    if(remain < 0) 
    {
        
        return -1;
    }

    return 0;
} 

unsigned long int WaveFile::getformtype(FILE* file, int filepos) {
    if(myfseek(file, filepos,SEEK_SET))
        return fourcc("");

    unsigned long int buf;
    if(sizeof(unsigned long int)!=myfread(&buf,1,sizeof(unsigned long int),file))
        return fourcc("");

    return buf;
} 

unsigned char imap[256] =
    {
        /*00*/ 0x00, 0x80, 0x40, 0xC0, 0x20, 0xA0, 0x60, 0xE0,
        /*08*/ 0x10, 0x90, 0x50, 0xD0, 0x30, 0xB0, 0x70, 0xF0,
        /*10*/ 0x08, 0x88, 0x48, 0xC8, 0x28, 0xA8, 0x68, 0xE8,
        /*18*/ 0x18, 0x98, 0x58, 0xD8, 0x38, 0xB8, 0x78, 0xF8,
        /*20*/ 0x04, 0x84, 0x44, 0xC4, 0x24, 0xA4, 0x64, 0xE4,
        /*28*/ 0x14, 0x94, 0x54, 0xD4, 0x34, 0xB4, 0x74, 0xF4,
        /*30*/ 0x0C, 0x8C, 0x4C, 0xCC, 0x2C, 0xAC, 0x6C, 0xEC,
        /*38*/ 0x1C, 0x9C, 0x5C, 0xDC, 0x3C, 0xBC, 0x7C, 0xFC,
        /*40*/ 0x02, 0x82, 0x42, 0xC2, 0x22, 0xA2, 0x62, 0xE2,
        /*48*/ 0x12, 0x92, 0x52, 0xD2, 0x32, 0xB2, 0x72, 0xF2,
        /*50*/ 0x0A, 0x8A, 0x4A, 0xCA, 0x2A, 0xAA, 0x6A, 0xEA,
        /*58*/ 0x1A, 0x9A, 0x5A, 0xDA, 0x3A, 0xBA, 0x7A, 0xFA,
        /*60*/ 0x06, 0x86, 0x46, 0xC6, 0x26, 0xA6, 0x66, 0xE6,
        /*68*/ 0x16, 0x96, 0x56, 0xD6, 0x36, 0xB6, 0x76, 0xF6,
        /*70*/ 0x0E, 0x8E, 0x4E, 0xCE, 0x2E, 0xAE, 0x6E, 0xEE,
        /*78*/ 0x1E, 0x9E, 0x5E, 0xDE, 0x3E, 0xBE, 0x7E, 0xFE,
        /*80*/ 0x01, 0x81, 0x41, 0xC1, 0x21, 0xA1, 0x61, 0xE1,
        /*88*/ 0x11, 0x91, 0x51, 0xD1, 0x31, 0xB1, 0x71, 0xF1,
        /*90*/ 0x09, 0x89, 0x49, 0xC9, 0x29, 0xA9, 0x69, 0xE9,
        /*98*/ 0x19, 0x99, 0x59, 0xD9, 0x39, 0xB9, 0x79, 0xF9,
        /*A0*/ 0x05, 0x85, 0x45, 0xC5, 0x25, 0xA5, 0x65, 0xE5,
        /*A8*/ 0x15, 0x95, 0x55, 0xD5, 0x35, 0xB5, 0x75, 0xF5,
        /*B0*/ 0x0D, 0x8D, 0x4D, 0xCD, 0x2D, 0xAD, 0x6D, 0xED,
        /*B8*/ 0x1D, 0x9D, 0x5D, 0xDD, 0x3D, 0xBD, 0x7D, 0xFD,
        /*C0*/ 0x03, 0x83, 0x43, 0xC3, 0x23, 0xA3, 0x63, 0xE3,
        /*C8*/ 0x13, 0x93, 0x53, 0xD3, 0x33, 0xB3, 0x73, 0xF3,
        /*D0*/ 0x0B, 0x8B, 0x4B, 0xCB, 0x2B, 0xAB, 0x6B, 0xEB,
        /*D8*/ 0x1B, 0x9B, 0x5B, 0xDB, 0x3B, 0xBB, 0x7B, 0xFB,
        /*E0*/ 0x07, 0x87, 0x47, 0xC7, 0x27, 0xA7, 0x67, 0xE7,
        /*E8*/ 0x17, 0x97, 0x57, 0xD7, 0x37, 0xB7, 0x77, 0xF7,
        /*F0*/ 0x0F, 0x8F, 0x4F, 0xCF, 0x2F, 0xAF, 0x6F, 0xEF,
        /*F8*/ 0x1F, 0x9F, 0x5F, 0xDF, 0x3F, 0xBF, 0x7F, 0xFF
    };

unsigned char*wav2law(unsigned char*wavedata, int len, int *len2) {
    WaveFile inSample(wavedata, len);
    int status=inSample.readHeaderAndFormat();
    if(status) {
        fprintf(stderr,"Error reading header");
        *len2 = 0;
        return 0;
    }
    int samplerate=inSample.getSamplesPerSec();
    int samplelen=inSample.filesamplelen();
    int bps=inSample.format.nBitsPerSample;
    

    int n_read;
    const int BLOCK=1024;
    unsigned char inbuf[BLOCK+8];
    unsigned char outbuf[BLOCK+8];

    double ratio = 8000.0/samplerate;
    double pos = 0;
    int pos2 = 0;
    unsigned char * newdata = (unsigned char*)malloc((int)(samplelen*ratio)+128);

    while((n_read=inSample.readDataBlock((char*)inbuf, BLOCK)) > 0) {
        pos -= (int)pos;

        if(bps == 8) {
            if(ratio <= 1) {
                for(int i=0; i<n_read; i++) {
                    unsigned char c = PCM16_TO_ALAW(((unsigned char)inbuf[i]-128)*256);
                    outbuf[(int)pos]=imap[c];
                    pos += ratio;
                }
            } else {
                for(int i=0; i<n_read; i++) {
                    unsigned char c = PCM16_TO_ALAW(((unsigned char)inbuf[i]-128)*256);
                    memset(&outbuf[(int)pos],imap[c], (int)ratio+1);
                    pos += ratio;
                }
            }
        } else if(bps == 16) {
            if(ratio <= 1) {
                for(int i=0; i<n_read/2; i++) {
                    unsigned char c = PCM16_TO_ALAW((short)(*(unsigned short*)&inbuf[i*2]));
                    outbuf[(int)pos]=imap[c];
                    pos += ratio;
                }
            } else {
                for(int i=0; i<n_read/2; i++) {
                    unsigned char c = PCM16_TO_ALAW((short)(*(unsigned short*)&inbuf[i*2]));
                    memset(&outbuf[(int)pos],imap[c], (int)ratio+1);
                    pos += ratio;
                }
            }
        }

        int num = (int)pos;
        memcpy(&newdata[pos2], outbuf, num);
        pos2 += num;
    }
    *len2 = pos2;
    return newdata;
}

int x_main(int argc, char* argv[]) {
    int nRetCode = 0;

    if(argc != 3) {
        fprintf(stderr, "Error: please specify 2 parameters\n");
        nRetCode = 1;
        return 1;
    }
    char *infilename=argv[1];
    char *outfilename=argv[2];

    FILE *outfile;

    WaveFile inSample(infilename);
    int status=inSample.readHeaderAndFormat();
    if(status) {
        fprintf(stderr,"Error reading header of %s\n",infilename);
        exit(1);
    }
    int samplerate=inSample.getSamplesPerSec();
    int samplelen=inSample.filesamplelen();
    int samplebytelen=samplelen*2;
    printf("samplerate: %d\n",samplerate);
    printf("samplelen: %d\n",samplelen);
    printf("samplebytelen: %d\n",samplebytelen);

    outfile=fopen(outfilename,"wb");

    const int BLOCK=1024;
    unsigned short inbuf[BLOCK];
    unsigned char outbuf[BLOCK];

    int n_read;
    int ii=0;
    while((n_read=inSample.readDataBlock((char*)inbuf, sizeof(inbuf[0])*BLOCK)) > 0) {
        for(int i=0; i<n_read/2; i++) {
            outbuf[i]=PCM16_TO_ALAW(inbuf[i]);
            outbuf[i] = imap[outbuf[i]];
            ii++;
        }

        fwrite(outbuf, sizeof(outbuf[0]), n_read/2, outfile);
        
    }
    printf("%d\n",ii);
    fclose(outfile);

    return 1;
}


